from __future__ import print_function

import numpy as np
from . import nat_cst as nc


def sigma_hm_ff(lambda_angstroem, temp, P_e):
    """
    Returns the H- free-free cross-section in units of cm^2
    per H per e- pressure (in cgs), as defined on page 156 of
    "The Observation and Analysis of Stellar Photospheres"
    by David F. Gray
    """

    index = (lambda_angstroem >= 2600.) & (lambda_angstroem <= 113900.)
    lamb_use = lambda_angstroem[index]

    if temp >= 2500.:
        # Convert to Angstrom (from cgs)
        theta = 5040. / temp

        f0 = -2.2763 - 1.6850 * np.log10(lamb_use) + \
          0.76661*np.log10(lamb_use)**2. \
          - 0.053346*np.log10(lamb_use)**3.
        f1 = 15.2827 - 9.2846 * np.log10(lamb_use) + \
          1.99381*np.log10(lamb_use)**2. \
          - 0.142631*np.log10(lamb_use)**3.
        f2 = -197.789 + 190.266 * np.log10(lamb_use) - 67.9775*np.log10(lamb_use)**2. \
                 + 10.6913*np.log10(lamb_use)**3. - 0.625151*np.log10(lamb_use)**4.

        ret_val = np.zeros_like(lambda_angstroem)
        ret_val[index] = 1e-26 * P_e * 1e1 ** (
                f0 + f1 * np.log10(theta) + f2 * np.log10(theta) ** 2.)
        return ret_val

    else:

        return np.zeros_like(lambda_angstroem)


def sigma_bf_mean(border_lambda_angstroem):
    """
    Returns the H- bound-free cross-section in units of cm^2 \
    per H-, as defined on page 155 of
    "The Observation and Analysis of Stellar Photospheres"
    by David F. Gray
    """

    left = border_lambda_angstroem[:-1]
    right = border_lambda_angstroem[1:]
    diff = np.diff(border_lambda_angstroem)

    a = [
        1.99654,
        -1.18267e-5,
        2.64243e-6,
        -4.40524e-10,
        3.23992e-14,
        -1.39568e-18,
        2.78701e-23
    ]

    ret_val = np.zeros_like(border_lambda_angstroem[1:])

    index = right <= 1.64e4

    for i_a in range(len(a)):
        ret_val[index] += a[i_a] * (
                right[index] ** (i_a + 1) - left[index] ** (i_a + 1)
        ) / (i_a + 1)

    index_bracket = (left < 1.64e4) & (right > 1.64e4)
    for i_a in range(len(a)):
        ret_val[index_bracket] += a[i_a] * (1.64e4 ** (i_a + 1) -
                                            left[index_bracket] ** (i_a + 1)) / (i_a + 1)

    index = (left + right) / 2. > 1.64e4
    ret_val[index] = 0.
    index = ret_val < 0.
    ret_val[index] = 0.

    return ret_val * 1e-18 / diff


def hminus_opacity(lambda_angstroem, border_lambda_angstroem,
                   temp, press, mmw, abundances):
    """ Calc the H- opacity."""

    ret_val = np.array(np.zeros(len(lambda_angstroem) * len(press)).reshape(
        len(lambda_angstroem),
        len(press)), dtype='d', order='F')

    # Calc. electron number fraction
    # e- mass in amu:
    m_e = 5.485799e-4
    n_e = mmw / m_e * abundances['e-']

    # Calc. e- partial pressure
    p_e = press * n_e

    kappa_hminus_bf = sigma_bf_mean(border_lambda_angstroem) / nc.amu

    for i_struct in range(len(n_e)):
        kappa_hminus_ff = sigma_hm_ff(lambda_angstroem, temp[i_struct],
                                      p_e[i_struct]) / nc.amu * abundances['H'][i_struct]

        ret_val[:, i_struct] = kappa_hminus_bf * abundances['H-'][i_struct] \
            + kappa_hminus_ff

    return ret_val


# Functions to read custom PT grids
# Function to sort custom (potentially randomly sorted) PT grid of opacities
def sort_opa_PTgrid(path_ptg):
    # Read the Ps and Ts
    p_ts = np.genfromtxt(path_ptg)

    # Read the file names
    f = open(path_ptg)
    lines = f.readlines()
    f.close()

    n_entries = len(lines)

    # Prepare the array to contain the
    # pressures, temperatures, indices in the unsorted list.
    # Also prepare the list of unsorted names
    p_tind = np.ones(n_entries * 3).reshape(n_entries, 3)
    names = []

    # Fill the array and name list
    for i_line in range(n_entries):

        line = lines[i_line]
        lsp = line.split(' ')

        p_tind[i_line, 0], p_tind[i_line, 1], p_tind[i_line, 2] = \
            p_ts[i_line, 0], p_ts[i_line, 1], i_line

        if lsp[-1][-1] == '\n':
            names.append(lsp[-1][:-1])
        else:
            names.append(lsp[-1])

    # Sort the array by temperature
    tsortind = np.argsort(p_tind[:, 1])
    p_tind = p_tind[tsortind, :]

    # Sort the array entries with constant
    # temperatures by pressure
    diff_ps = 0
    t_start = p_tind[0, 1]

    for i in range(n_entries):
        if np.abs(p_tind[i, 1] - t_start) > 1e-10:
            break
        diff_ps = diff_ps + 1

    diff_ts = int(n_entries / diff_ps)
    for i_dT in range(diff_ts):
        subsort = p_tind[i_dT * diff_ps:(i_dT + 1) * diff_ps, :]
        psortind = np.argsort(subsort[:, 0])
        subsort = subsort[psortind, :]
        p_tind[i_dT * diff_ps:(i_dT + 1) * diff_ps, :] = subsort

    names_sorted = []
    for i_line in range(n_entries):
        names_sorted.append(names[int(p_tind[i_line, 2] + 0.01)])

    # Convert from bars to cgs
    p_tind[:, 0] = p_tind[:, 0] * 1e6

    return [p_tind[:, :-1][:, ::-1], names_sorted, diff_ts, diff_ps]


# Check if custom grid exists, if yes return sorted P-T array with
# corresponding sorted path_input_data names, retutn None otherwise.

def get_custom_PT_grid(path, mode, species):
    import os as os

    path_test = path + '/opacities/lines/'
    if mode == 'lbl':
        path_test = path_test + 'line_by_line/'
    elif mode == 'c-k':
        path_test = path_test + 'corr_k/'
    path_test = path_test + species + '/PTpaths.ls'
    if not os.path.isfile(path_test):
        return None
    else:
        return sort_opa_PTgrid(path_test)


'''
import pylab as plt

demo_temp = 2880
demo_press = 0.33*1e6
ab_em = 1e-6*5.485799e-4/2.33
ab_h = 0.33*1./2.33
ab_hm = 2e-9*1./2.33
mww_val = 2.33

abunds = {}
abunds['e-'] = np.array([ab_em,ab_em,ab_em])
abunds['H'] = np.array([ab_h,ab_h,ab_h])
abunds['H-'] = np.array([ab_hm,ab_hm,ab_hm])

temp = np.array([demo_temp, demo_temp, demo_temp])
press = np.array([demo_press, demo_press, demo_press])
mmw = np.array([mww_val, mww_val, mww_val])

lamb = np.logspace(np.log10(0.9),np.log10(10.),7000)*1e4
lamb_coarse = np.logspace(np.log10(0.9),np.log10(10.),10)*1e4

def calc_borders(x):
        xn = []
        xn.append(x[0]-(x[1]-x[0])/2.)
        for i in range(int(len(x))-1):
            xn.append(x[i]+(x[i+1]-x[i])/2.)
        xn.append(x[int(len(x))-1]+(x[int(len(x))-1]-x[int(len(x))-2])/2.)
        return np.array(xn)

lamb_bord = calc_borders(lamb)
print('a')
for i in range(33):
    opa = hminus_opacity(lamb, lamb_bord, temp, press, mmw, abunds)
print('b')

plt.plot(lamb/1e4, opa*2.33*nc.amu)

lamb_coarse_bord = calc_borders(lamb_coarse)
opa = hminus_opacity(lamb_coarse, lamb_coarse_bord, temp, press, mmw, abunds)
plt.plot(lamb_coarse/1e4, opa*2.33*nc.amu)

plt.ylim([1e-28,1e-22])
plt.xscale('log')
plt.yscale('log')
plt.show()
'''



"""Test the radtrans module in line-by-line mode with sampling.

Do not change the parameters used to generate the comparison files, including input_data files, when running the tests.
"""
from .context import petitRADTRANS
from .utils import compare_from_reference_file, \
    reference_filenames, radtrans_parameters, temperature_guillot_2010, temperature_isothermal

relative_tolerance = 1e-6  # relative tolerance when comparing with older results


# Initializations
def init_radtrans_downsampled_line_by_line():
    atmosphere = petitRADTRANS.radtrans.Radtrans(
        line_species=radtrans_parameters['spectrum_parameters']['line_species_line_by_line'],
        rayleigh_species=radtrans_parameters['spectrum_parameters']['rayleigh_species'],
        continuum_opacities=radtrans_parameters['spectrum_parameters']['continuum_opacities'],
        wlen_bords_micron=radtrans_parameters['spectrum_parameters']['wavelength_range_line_by_line'],
        lbl_opacity_sampling=radtrans_parameters['spectrum_parameters']['line_by_line_opacity_sampling'],
        mode='lbl'
    )

    atmosphere.setup_opa_structure(radtrans_parameters['pressures'])

    return atmosphere


atmosphere_lbl_downsampled = init_radtrans_downsampled_line_by_line()


def test_line_by_line_downsampled_emission_spectrum():
    # Calculate an emission spectrum
    atmosphere_lbl_downsampled.calc_flux(
        temp=temperature_guillot_2010,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['line_by_line_downsampled_emission'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_lbl_downsampled.freq * 1e4,
            'spectral_radiosity': atmosphere_lbl_downsampled.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_line_by_line_downsampled_transmission_spectrum():
    # Calculate a transmission spectrum
    atmosphere_lbl_downsampled.calc_transm(
        temp=temperature_isothermal,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['line_by_line_downsampled_transmission'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_lbl_downsampled.freq * 1e4,
            'transit_radius': atmosphere_lbl_downsampled.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )



import sys
import os
import logging
# To not have numpy start parallelizing on its own
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
from .data import Data
from .parameter import Parameter

#import species
#species.SpeciesInit()

class RetrievalConfig:
    """
    The RetrievalConfig class contains all of the data and model level information necessary
    to run a petitRADTRANS retrieval. The name of the class will be used to name outputs.
    This class is passed to the Retrieval, which runs the actual pymultinest retrieval
    and produces the outputs.

    The general usage of this class is to define it, add the parameters and their priors,
    add the opacity sources, the data together with a model for each dataset, and then
    configure a few plotting arguments.

    Args:
        retrieval_name : str
            Name of this retrieval. Make it informative so that you can keep track of the outputs!
        run_mode : str
            Can be either 'retrieval', which runs the retrieval normally using pymultinest,
            or 'evaluate', which produces plots from the best fit parameters stored in the
            output post_equal_weights file.
        AMR : bool
            Use an adaptive high resolution pressure grid around the location of cloud condensation.
            This will increase the size of the pressure grid by a constant factor that can be adjusted
            in the setup_pres function.
        scattering : bool
            If using emission spectra, turn scattering on or off.
        pressures : numpy.array
            A log-spaced array of pressures over which to retrieve. 100 points is standard, between
            10^-6 and 10^3.
    """

    def __init__(self,
                 retrieval_name="retrieval_name",
                 run_mode="retrieval",
                 AMR=False,
                 scattering=False,
                 distribution="lognormal",
                 pressures=None,
                 write_out_spec_sample=False):

        self.retrieval_name = retrieval_name

        if run_mode == 'retrieve':
            run_mode = 'retrieval'  # TODO please no.

        self.run_mode = run_mode

        if self.run_mode != 'retrieval' and self.run_mode != 'evaluate':
            logging.error("run_mode must be either 'retrieval' or 'evaluate'!")
            sys.exit(1)
        self.AMR = AMR

        if pressures is not None:
            self.p_global = pressures
        else:
            self.p_global = np.logspace(-6, 3, 100)

        self.scattering = scattering
        self.distribution = distribution
        self.parameters = {}  #: Dictionary of the parameters passed to the model generating function
        self.data = {}  #: Dictionary of the datasets used in the retrieval.
        self.instruments = []
        self.line_species = []
        self.cloud_species = []
        self.rayleigh_species = []
        self.continuum_opacities = []
        self.plot_kwargs = {}

        self._plot_defaults()
        self.write_out_spec_sample = write_out_spec_sample

        self.add_parameter("pressure_scaling", False, value=1)
        self.add_parameter("pressure_width", False, value=1)
        self.add_parameter("pressure_simple", False, value=self.p_global.shape[0])

    def _plot_defaults(self):
        ##################################################################
        # Define axis properties of spectral plot if run_mode == 'evaluate'
        ##################################################################
        self.plot_kwargs["spec_xlabel"] = 'Wavelength [micron]'
        self.plot_kwargs["spec_ylabel"] = "Flux [W/m2/micron]"
        self.plot_kwargs["y_axis_scaling"] = 1.0
        self.plot_kwargs["xscale"] = 'log'
        self.plot_kwargs["yscale"] = 'linear'
        self.plot_kwargs["resolution"] = 1500.
        self.plot_kwargs["nsample"] = 10.

        ##################################################################
        # Define from which observation object to take P-T
        # in evaluation mode (if run_mode == 'evaluate'),
        # add PT-envelope plotting options
        ##################################################################
        self.plot_kwargs["temp_limits"] = [150, 3000]
        self.plot_kwargs["press_limits"] = [1e2, 1e-5]

    def _setup_pres(self, scaling=10, width=3):
        """
        This converts the standard pressure grid into the correct length
        for the AMR pressure grid. The scaling adjusts the resolution of the
        high resolution grid, while the width determines the size of the high
        pressure region. This function is automatically called in
        Retrieval.setupData().

        Args:
            scaling : int
                A multiplicative factor that determines the size of the full high resolution pressure grid,
                which will have length self.p_global.shape[0] * scaling.
            width : int
                The number of cells in the low pressure grid to replace with the high resolution grid.
        """

        print("Setting up AMR pressure grid.")
        self.scaling = scaling
        self.width = width
        nclouds = len(self.cloud_species)
        if nclouds == 0:
            print("WARNING: there are no clouds in the retrieval, please add cloud species before setting up amr")
        new_len = self.p_global.shape[0] + nclouds * width * (scaling - 1)
        self.amr_pressure = np.logspace(np.log10(self.p_global[0]), np.log10(self.p_global[-1]), new_len)
        self.add_parameter("pressure_scaling", False, value=scaling)
        self.add_parameter("pressure_width", False, value=width)
        self.add_parameter("pressure_simple", False, value=self.p_global.shape[0])

        return self.amr_pressure

    def add_parameter(self, name, free, value=None, transform_prior_cube_coordinate=None):
        """
        This function adds a Parameter (see parameter.py) to the dictionary of parameters. A Parameter
        has a name and a boolean parameter to set whether it is a free or fixed parameter during the retrieval.
        In addition, a value can be set, or a prior function can be given that transforms a random variable in
        [0,1] to the physical dimensions of the Parameter.

        Args:
            name : str
                The name of the parameter. Must match the name used in the model function for the retrieval.
            free : bool
                True if the parameter is a free parameter in the retrieval, false if it is fixed.
            value : float
                The value of the parameter in the units used by the model function.
            transform_prior_cube_coordinate : method
                A function that transforms the unit interval to the physical units of the parameter.
                Typically given as a lambda function.
        """

        self.parameters[name] = Parameter(name, free, value,
                                          transform_prior_cube_coordinate=transform_prior_cube_coordinate)

    def list_available_line_species(self):  # TODO could be static
        """
        List the currently installed opacity tables that are available for species that contribute to the line opacity.
        """

        prt_path = os.environ.get("pRT_input_data_path")
        if prt_path is None:
            raise OSError(f"Path to input data not specified!\n"
                          f"Please set pRT_input_data_path variable in .bashrc / .bash_profile or specify path via\n"
                          f">>> import os"
                          f">>> os.environ['pRT_input_data_path'] = 'absolute/path/of/the/folder/input_data'\n"
                          f"before creating a Radtrans object or loading the nat_cst module.\n"
                          f"(this will become unnecessary in a future update)"
                          )

        files = [f[0].split('/')[-1] for f in os.walk(prt_path + "/opacities/lines/corr_k/")]
        files = set([f.split('_R_')[0] for f in files])
        print("\ncorrelated-k opacities")

        for f in files:
            print(f)

        lbl = [f[0].split('/')[-1] for f in os.walk(prt_path + "/opacities/lines/line_by_line/")]
        lbl = set(lbl)
        print("\nline-by-line opacities")

        for f in lbl:
            print(f)

        return files.union(lbl)

    def list_available_cloud_species(self):  # TODO could be static
        """
        List the currently installed opacity tables that are available for cloud species.
        """

        prt_path = os.environ.get("pRT_input_data_path")
        if prt_path is None:
            raise OSError(f"Path to input data not specified!\n"
                          f"Please set pRT_input_data_path variable in .bashrc / .bash_profile or specify path via\n"
                          f">>> import os"
                          f">>> os.environ['pRT_input_data_path'] = 'absolute/path/of/the/folder/input_data'\n"
                          f"before creating a Radtrans object or loading the nat_cst module.\n"
                          f"(this will become unnecessary in a future update)"
                          )

        files = [f[0].split('/')[-1] for f in os.walk(prt_path + "/opacities/continuum/clouds/")]
        files = set(files)

        for f in files:
            print(f)

        return files

    def list_available_cia_species(self):  # TODO could be static
        """
        List the currently installed opacity tables that are available for CIA species.
        """

        prt_path = os.environ.get("pRT_input_data_path")
        if prt_path is None:
            raise OSError(f"Path to input data not specified!\n"
                          f"Please set pRT_input_data_path variable in .bashrc / .bash_profile or specify path via\n"
                          f">>> import os"
                          f">>> os.environ['pRT_input_data_path'] = 'absolute/path/of/the/folder/input_data'\n"
                          f"before creating a Radtrans object or loading the nat_cst module.\n"
                          f"(this will become unnecessary in a future update)"
                          )

        files = [f[0].split('/')[-1] for f in os.walk(prt_path + "/opacities/continuum/cia/")]
        files = set(files)

        for f in files:
            print(f)

        return files

    def set_line_species(self, linelist, eq=False, abund_lim=(-6.0, -0.5)):
        """
        Set RadTrans.line_species

        This function adds a list of species to the pRT object that will define the line
        opacities of the model. The values in the list are strings, with the names matching
        the pRT opacity names, which vary between the c-k line opacities and the line-by-line opacities.

        NOTE: As of pRT version 2.4.9, the behaviour of this function has changed. In previous versions the
        abundance limits were set from abund_lim[0] to (abund_lim[0] + abund_lim[1]). This has been changed
        so that the limits of the prior range are from abund_lim[0] to abund_lim[1] (ie the actual boundaries).

        Args:
            linelist : List(str)
                The list of species to include in the retrieval
            eq : bool
                If false, the retrieval should use free chemistry, and Parameters for the abundance of each
                species in the linelist will be added to the retrieval. Otherwise, equilibrium chemistry will
                be used. If you need fine control species, use the add_line_species and set up each species
                individually.
            abund_lim : Tuple(float,float)
                If free is True, this sets the boundaries of the uniform prior that will be applied for
                each species in linelist. The range of the prior goes from abund_lim[0] to abund_lim[1].
                The abundance limits must be given in log10 units of the mass fraction.
        """
        if abund_lim[1] > 0.0:
            raise ValueError(f"upper limit must be <= 0.0 (was {abund_lim})! Please set abundance limits as (low, high)")

        self.line_species = linelist

        if not eq:
            for spec in self.line_species:
                self.parameters[spec] = Parameter(
                    spec,
                    True,
                    transform_prior_cube_coordinate=lambda x: abund_lim[0] + (abund_lim[1] - abund_lim[0]) * x
                )

    def set_rayleigh_species(self, linelist):
        """
        Set the list of species that contribute to the rayleigh scattering in the pRT object.

        Args:
            linelist : List(str)
                A list of species that contribute to the rayleigh opacity.
        """

        self.rayleigh_species = linelist

    def set_continuum_opacities(self, linelist):
        """
        Set the list of species that contribute to the continuum opacity in the pRT object.

        Args:
            linelist : List(str)
                A list of species that contribute to the continuum opacity.
        """

        self.continuum_opacities = linelist

    def add_line_species(self, species, eq=False, abund_lim=(-7.0, 0.0), fixed_abund=None):
        """
        This function adds a single species to the pRT object that will define the line opacities of the model.
        The name must match the pRT opacity name, which vary between the c-k line opacities and the line-by-line
        opacities.

        NOTE: As of pRT version 2.4.9, the behaviour of this function has changed. In previous versions the
        abundance limits were set from abund_lim[0] to (abund_lim[0] + abund_lim[1]). This has been changed
        so that the limits of the prior range are from abund_lim[0] to abund_lim[1] (ie the actual boundaries).

        Args:
            species : str
                The species to include in the retrieval
            eq : bool
                If False, the retrieval should use free chemistry, and Parameters for the abundance of the
                species will be added to the retrieval. Otherwise, (dis)equilibrium chemistry will be used.
            abund_lim : Tuple(float,float)
                If free is True, this sets the boundaries of the uniform prior that will be applied the species given.
                The range of the prior goes from abund_lim[0] to abund_lim[1]
                The abundance limits must be given in log10 units of the mass fraction.
            fixed_abund : float
                The log-mass fraction abundance of the species. Currently only supports vertically constant
                abundances. If this is set, then the species will not be a free parameter in the retrieval.
        """

        # parameter passed through loglike is log10 abundance
        if abund_lim[1] > 0.0:
            raise ValueError(f"upper limit must be <= 0.0 (was {abund_lim})! Please set abundance limits as (low, high)")

        self.line_species.append(species)
        if not eq:
            if fixed_abund is not None:
                self.parameters[species] = Parameter(species, False,
                                                     value=fixed_abund)
            else:
                self.parameters[species] = Parameter(
                    species,
                    True,
                    transform_prior_cube_coordinate=lambda x: abund_lim[0] + (abund_lim[1] - abund_lim[0]) * x
                )

    def remove_species_lines(self, species, free=False):
        """
        This function removes a species from the pRT line list, and if using a free chemistry retrieval,
        removes the associated Parameter of the species.

        Args:
            species : str
                The species to remove from the retrieval
            free : bool
                If true, the retrieval should use free chemistry, and Parameters for the abundance of the
                species will be removed to the retrieval
        """

        if species in self.line_species:
            self.line_species.remove(species)
        if free:
            self.parameters.pop(species, None)

    def add_cloud_species(self,species, eq = True, abund_lim = (-3.5,1.5), scaling_factor = None, 
                          PBase_lim = None, fixed_abund = None, fixed_base=None):
        """
        This function adds a single cloud species to the list of species. Optionally,
        it will add parameters to allow for a retrieval using an ackermann-marley model.
        If an equilibrium condensation model is used in th retrieval model function (eq=True),
        then a parameter is added that scales the equilibrium cloud abundance, as in Molliere (2020).
        If eq is false, two parameters are added, the cloud abundnace and the cloud base pressure.
        The limits set the prior ranges, both on a log scale.

        NOTE: As of pRT version 2.4.9, the behaviour of this function has changed. In previous versions the
        abundance limits were set from abund_lim[0] to (abund_lim[0] + abund_lim[1]). This has been changed
        so that the limits of the prior range are from abund_lim[0] to abund_lim[1] (ie the actual boundaries).
        The same is true for PBase_lim.

        Args:
            species : str
                Name of the pRT cloud species, including the cloud shape tag.
            eq : bool
                Does the retrieval model use an equilibrium cloud model. This restricts the available species!
            abund_lim : tuple(float,float)
                If eq is True, this sets the scaling factor for the equilibrium condensate abundance, typical
                range would be (-3,1). If eq is false, this sets the range on the actual cloud abundance,
                with a typical range being (-5,0).
            PBase_lim : tuple(float,float)
                Only used if not using an equilibrium model. Sets the limits on the log of the cloud base pressure.
                Obsolete.
            fixed_abund : Optional(float)
                A vertically constant log mass fraction abundance for the cloud species. If set, this will not be
                a free parameter in the retrieval. Only compatible with non-equilibrium clouds.
            fixed_base : Optional(float)
                The log cloud base pressure. If set, fixes this parameter to a constant value, and it will not be
                a free parameter in the retrieval. Only compatible with non-equilibrium clouds. Not yet compatible
                with most built in pRT models.
        """

        if species.endswith("(c)"):
            logging.warning("Ensure you set the cloud particle shape, typically with the _cd tag!")
            logging.warning(species + " was not added to the list of cloud species")
            return

        self.cloud_species.append(species)
        cname = species.split('_')[0]
        if scaling_factor is not None:
            self.parameters['eq_scaling_'+cname] = Parameter('eq_scaling_'+cname,True,\
                                                transform_prior_cube_coordinate = \
                                                lambda x : scaling_factor[0] + (scaling_factor[1]-scaling_factor[0])*x)
        if not eq:
            if abund_lim[1] > 0.0:
                raise ValueError(
                    f"upper limit must be <= 0.0 (was {abund_lim})! Please set abundance limits as (low, high)"
                )

            if fixed_abund is None:
                self.parameters['log_X_cb_' + cname] = Parameter(
                    'log_X_cb_' + cname,
                    True,
                    transform_prior_cube_coordinate=lambda x: abund_lim[0] + (abund_lim[1] - abund_lim[0]) * x
                )
            else:
                self.parameters['log_X_cb_' + cname] = Parameter(
                    'log_X_cb_' + cname,
                    False,
                    value=fixed_abund
                )

        if PBase_lim is not None or fixed_base is not None:
            if fixed_base is None:
                self.parameters['log_Pbase_' + cname] = Parameter(
                    'log_Pbase_' + cname,
                    True,
                    transform_prior_cube_coordinate=lambda x: PBase_lim[0] + (PBase_lim[1] - PBase_lim[0]) * x
                )
            else:
                self.parameters['log_Pbase_' + cname] = Parameter(
                    'log_Pbase_' + cname,
                    False,
                    value=fixed_base
                )

    def add_data(self, 
                 name, 
                 path,
                 model_generating_function,
                 data_resolution = None,
                 model_resolution = None,
                 distance = None,
                 scale = False,
                 scale_err = False,
                 offset_bool = False,
                 wlen_range_micron = None,
                 external_pRT_reference = None,
                 opacity_mode = 'c-k',
                 wlen_bins=None,
                 pRT_grid=False,
                 pRT_object=None,
                 wlen=None,
                 flux=None,
                 flux_error=None,
                 mask=None):
        """
        Create a Data class object.
        # TODO complete docstring
        Args:
            name : str
                Identifier for this data set.
            path : str
                Path to observations file, including filename. This can be a txt or dat file containing the wavelength,
                flux, transit depth and error, or a fits file containing the wavelength, spectrum and covariance matrix.
            model_generating_function : fnc
                A function, typically defined in run_definition.py that returns the model wavelength and
                spectrum (emission or transmission). This is the function that contains the physics
                of the model, and calls pRT in order to compute the spectrum.
            data_resolution : float
                Spectral resolution of the instrument. Optional, allows convolution of model to instrumental line width.
            model_resolution : float
                Spectral resolution of the model, allowing for low resolution correlated k tables from exo-k.
            distance : float
                The distance to the object in cgs units. Defaults to a 10pc normalized distance. All data must
                be scaled to the same distance before running the retrieval, which can be done using the
                scale_to_distance method in the Data class.
            scale : bool
                Turn on or off scaling the data by a constant factor.
            wlen_range_micron : Tuple
                A pair of wavelenths in units of micron that determine the lower and upper boundaries of the
                model computation.
            external_pRT_reference : str
                The name of an existing Data object. This object's prt_object will be used to calculate the chi squared
                of the new Data object. This is useful when two datasets overlap, as only one model computation is
                required to compute the log likelihood of both datasets.
            opacity_mode : str
                Should the retrieval be run using correlated-k opacities (default, 'c-k'),
                or line by line ('lbl') opacities? If 'lbl' is selected, it is HIGHLY
                recommended to set the model_resolution parameter.
            pRT_grid: bool
                Set to true if data has been binned to pRT R = 1,000 c-k grid.
        """
        self.data[name] = Data(name, path,
                               model_generating_function=model_generating_function,
                               data_resolution=data_resolution,
                               model_resolution=model_resolution,
                               distance=distance,
                               scale=scale,
                               scale_err=scale_err,
                               offset_bool = offset_bool,
                               wlen_range_micron=wlen_range_micron,
                               external_pRT_reference=external_pRT_reference,
                               opacity_mode=opacity_mode,
                               wlen_bins=wlen_bins,
                               pRT_grid=pRT_grid,
                               pRT_object=pRT_object,
                               wlen=wlen,
                               flux=flux,
                               flux_error=flux_error,
                               mask=mask
                               )

    def add_photometry(self, path,
                       model_generating_function,
                       model_resolution=10,
                       distance=None,
                       scale=False,
                       wlen_range_micron=None,
                       photometric_transformation_function=None,
                       external_pRT_reference=None,
                       opacity_mode='c-k'):
        """
        Create a Data class object for each photometric point in a photometry file.
        The photometry file must be a csv file and have the following structure:
        name, lower wavelength bound [um], upper wavelength boundary[um], flux [W/m2/micron], flux error [W/m2/micron]

        Photometric data requires a transformation function to convert a spectrum into synthetic photometry.
        You must provide this function yourself, or have the species package installed.
        If using species, the name in the data file must be of the format instrument/filter.

        Args:
            model_generating_function : str
                Identifier for this data set.
            path : str
                Path to observations file, including filename.
            model_resolution : float
                Spectral resolution of the model, allowing for low resolution correlated k tables from exo-k.
            scale : bool
                Turn on or off scaling the data by a constant factor. Currently only set up to scale all
                photometric data in a given file.
            distance : float
                The distance to the object in cgs units. Defaults to a 10pc normalized distance. All data must
                be scaled to the same distance before running the retrieval, which can be done using the
                scale_to_distance method in the Data class.
            wlen_range_micron : Tuple
                A pair of wavelenths in units of micron that determine the lower and upper boundaries of
                the model computation.
            external_pRT_reference : str
                The name of an existing Data object. This object's prt_object will be used to calculate the
                chi squared of the new Data object. This is useful when two datasets overlap, as only
                one model computation is required to compute the log likelihood of both datasets.
            photometric_transformation_function : method
                A function that will transform a spectrum into an average synthetic photometric point,
                typicall accounting for filter transmission.
            opacity_mode: str
                Opacity mode.
        """

        with open(path) as photometry:
            if photometric_transformation_function is None:
                try:
                    import species as sp
                    sp.SpeciesInit()
                except ModuleNotFoundError:  # TODO find what error is expected here
                    logging.error(
                        "Please provide a function to transform a spectrum into photometry, or pip install species"
                    )
                    
            for line in photometry:
                # # must be the comment character
                if line[0] == '#':
                    continue

                vals = line.split(',')
                name = vals[0]
                wlow = float(vals[1])
                whigh = float(vals[2])
                flux = float(vals[3])
                err = float(vals[4])

                if photometric_transformation_function is None:
                    transform = sp.SyntheticPhotometry(name).spectrum_to_flux
                else:
                    transform = photometric_transformation_function

                if wlen_range_micron is None:
                    wbins = [0.95 * wlow, 1.05 * whigh]
                else:
                    wbins = wlen_range_micron

                if opacity_mode == 'lbl':
                    logging.warning("Are you sure you want a high resolution model for photometry?")

                self.data[name] = Data(
                    name,
                    path,
                    model_generating_function=model_generating_function,
                    distance=distance,
                    photometry=True,
                    wlen_range_micron=wbins,
                    photometric_bin_edges=[wlow, whigh],
                    data_resolution=np.mean([wlow, whigh]) / (whigh - wlow),
                    model_resolution=model_resolution,
                    scale=scale,
                    photometric_transformation_function=transform,
                    external_pRT_reference=external_pRT_reference,
                    opacity_mode=opacity_mode
                )
                self.data[name].flux = flux
                self.data[name].flux_error = err



"""Test the radtrans module.

Essentially go through a simplified version of the tutorial, and compare the results with previous ones.
C.f. (https://petitradtrans.readthedocs.io/en/latest/content/notebooks/getting_started.html).

Do not change the parameters used to generate the comparison files, including input_data files, when running the tests.
"""
import copy

import numpy as np

from .context import petitRADTRANS
from .utils import compare_from_reference_file, \
    reference_filenames, radtrans_parameters, temperature_guillot_2010, temperature_isothermal

relative_tolerance = 1e-6  # relative tolerance when comparing with older results


# Initializations
def init_radtrans_correlated_k():
    atmosphere = petitRADTRANS.radtrans.Radtrans(
        line_species=radtrans_parameters['spectrum_parameters']['line_species_correlated_k'],
        rayleigh_species=radtrans_parameters['spectrum_parameters']['rayleigh_species'],
        continuum_opacities=radtrans_parameters['spectrum_parameters']['continuum_opacities'],
        cloud_species=list(radtrans_parameters['cloud_parameters']['cloud_species'].keys()),
        wlen_bords_micron=radtrans_parameters['spectrum_parameters']['wavelength_range_correlated_k'],
        mode='c-k'
    )

    atmosphere.setup_opa_structure(radtrans_parameters['pressures'])

    return atmosphere


atmosphere_ck = init_radtrans_correlated_k()


# Tests
def test_correlated_k_emission_spectrum():
    # Calculate an emission spectrum
    atmosphere_ck.calc_flux(
        temp=temperature_guillot_2010,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_emission'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'spectral_radiosity': atmosphere_ck.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_emission_contribution_cloud_calculated_radius():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck.calc_flux(
        temp=temperature_guillot_2010,
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal'],
        contribution=True
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_emission_contribution_cloud_calculated_radius'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'contribution': atmosphere_ck.contr_em
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_emission_cloud_calculated_radius():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck.calc_flux(
        temp=temperature_guillot_2010,
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_emission_cloud_calculated_radius'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'spectral_radiosity': atmosphere_ck.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_emission_spectrum_cloud_hansen_radius():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck.calc_flux(
        temp=temperature_guillot_2010,
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        b_hans=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['b_hansen'],
        dist='hansen'
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_emission_cloud_hansen_radius'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'spectral_radiosity': atmosphere_ck.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_transmission_spectrum():
    # Calculate a transmission spectrum
    atmosphere_ck.calc_transm(
        temp=temperature_isothermal,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_transmission'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'transit_radius': atmosphere_ck.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_transmission_spectrum_cloud_power_law():
    # Calculate a transmission spectrum
    atmosphere_ck.calc_transm(
        temp=temperature_isothermal,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        kappa_zero=radtrans_parameters['cloud_parameters']['kappa_zero'],
        gamma_scat=radtrans_parameters['cloud_parameters']['gamma_scattering']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_transmission_cloud_power_law'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'transit_radius': atmosphere_ck.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_transmission_spectrum_gray_cloud():
    # Calculate a transmission spectrum
    atmosphere_ck.calc_transm(
        temp=temperature_isothermal,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        Pcloud=radtrans_parameters['cloud_parameters']['cloud_pressure']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_transmission_gray_cloud'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'transit_radius': atmosphere_ck.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_transmission_spectrum_rayleigh():
    # Calculate a transmission spectrum
    atmosphere_ck.calc_transm(
        temp=temperature_isothermal,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        haze_factor=radtrans_parameters['cloud_parameters']['haze_factor']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_transmission_rayleigh'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'transit_radius': atmosphere_ck.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_transmission_spectrum_cloud_fixed_radius():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck.calc_transm(
        temp=radtrans_parameters['temperature_isothermal'] * np.ones_like(radtrans_parameters['pressures']),
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        radius={'Mg2SiO4(c)': radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['radius']},
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_transmission_cloud_fixed_radius'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'transit_radius': atmosphere_ck.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_transmission_contribution_cloud_calculated_radius():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck.calc_transm(
        temp=radtrans_parameters['temperature_isothermal'] * np.ones_like(radtrans_parameters['pressures']),
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal'],
        contribution=True
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames[
            'correlated_k_transmission_contribution_cloud_calculated_radius'
        ],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'contribution': atmosphere_ck.contr_tr
        },
        relative_tolerance=relative_tolerance,
        absolute_tolerance=5e-10  # there is a max absolute error of 3.28989835374216e-10 between windows and WSL
                                  # generated files TODO investigate why
    )


def test_correlated_k_transmission_spectrum_cloud_calculated_radius():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck.calc_transm(
        temp=radtrans_parameters['temperature_isothermal'] * np.ones_like(radtrans_parameters['pressures']),
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_transmission_cloud_calculated_radius'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck.freq * 1e4,
            'transit_radius': atmosphere_ck.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def __test_plot_opacities():
    atmosphere_ck.plot_opas(
        species=[
            'H2O_HITEMP',
            'CH4'
        ],
        temperature=1500,
        pressure_bar=0.1
    )



"""Test the fortran modules.
"""
import numpy as np

from .context import petitRADTRANS


def test_rebin_spectrum():
    input_wavelengths = np.linspace(0.85, 2.15, 20)
    input_flux = np.sin(3 * input_wavelengths) + np.sin(10 * input_wavelengths)
    output_wavelengths = np.linspace(1, 2, 5)

    output_flux = petitRADTRANS.fort_rebin.rebin_spectrum(
            input_wavelengths, input_flux, output_wavelengths
    )

    assert np.allclose(output_flux, np.array([-0.26055379, -0.60538758, -0.47629003, -1.54747344,  0.3938092]))



"""Test the correlated-k scattering part of the radtrans module.

Essentially go through a simplified version of the tutorial, and compare the results with previous ones.
C.f. (https://petitradtrans.readthedocs.io/en/latest/content/notebooks/getting_started.html).

Do not change the parameters used to generate the comparison files, including input_data files, when running the tests.

Due to the way scattering and correlated-k are calculated in petitRADTRANS, results using the same parameters may have
variations of <~ 1%. To take that into account, an important relative tolerance is set for the tests, and multiple tests
may be performed in order to rule out "unlucky" results.
"""
import copy

import numpy as np

from .context import petitRADTRANS
from .utils import compare_from_reference_file, \
    reference_filenames, radtrans_parameters, temperature_guillot_2010

relative_tolerance = 1e-6  # relative tolerance when comparing with older spectra


# Initializations
def init_radtrans_correlated_k():
    atmosphere = petitRADTRANS.radtrans.Radtrans(
        line_species=radtrans_parameters['spectrum_parameters']['line_species_correlated_k'],
        rayleigh_species=radtrans_parameters['spectrum_parameters']['rayleigh_species'],
        continuum_opacities=radtrans_parameters['spectrum_parameters']['continuum_opacities'],
        cloud_species=list(radtrans_parameters['cloud_parameters']['cloud_species'].keys()),
        wlen_bords_micron=radtrans_parameters['spectrum_parameters']['wavelength_range_correlated_k'],
        mode='c-k',
        do_scat_emis=True
    )

    atmosphere.setup_opa_structure(radtrans_parameters['pressures'])

    return atmosphere


atmosphere_ck_scattering = init_radtrans_correlated_k()


def test_correlated_k_emission_spectrum_cloud_calculated_radius_scattering():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck_scattering.calc_flux(
        temp=temperature_guillot_2010,
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal'],
        add_cloud_scat_as_abs=True
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_emission_cloud_calculated_radius_scattering'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck_scattering.freq * 1e4,
            'spectral_radiosity': atmosphere_ck_scattering.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_emission_spectrum_cloud_calculated_radius_stellar_scattering_planetary_average():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    geometry = 'planetary_ave'

    atmosphere_ck_scattering.calc_flux(
        temp=temperature_guillot_2010,
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal'],
        geometry=geometry,
        Tstar=radtrans_parameters['stellar_parameters']['effective_temperature'],
        Rstar=radtrans_parameters['stellar_parameters']['radius'] * petitRADTRANS.nat_cst.r_sun,
        semimajoraxis=radtrans_parameters['planetary_parameters']['orbit_semi_major_axis']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames[
            'correlated_k_emission_cloud_calculated_radius_scattering_planetary_ave'
        ],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck_scattering.freq * 1e4,
            'spectral_radiosity': atmosphere_ck_scattering.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_emission_spectrum_cloud_calculated_radius_stellar_scattering_dayside():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    geometry = 'dayside_ave'

    atmosphere_ck_scattering.calc_flux(
        temp=temperature_guillot_2010,
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal'],
        geometry=geometry,
        Tstar=radtrans_parameters['stellar_parameters']['effective_temperature'],
        Rstar=radtrans_parameters['stellar_parameters']['radius'] * petitRADTRANS.nat_cst.r_sun,
        semimajoraxis=radtrans_parameters['planetary_parameters']['orbit_semi_major_axis']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_emission_cloud_calculated_radius_scattering_dayside_ave'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck_scattering.freq * 1e4,
            'spectral_radiosity': atmosphere_ck_scattering.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_emission_spectrum_cloud_calculated_radius_stellar_scattering_non_isotropic():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    geometry = 'non-isotropic'

    atmosphere_ck_scattering.calc_flux(
        temp=temperature_guillot_2010,
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal'],
        geometry=geometry,
        Tstar=radtrans_parameters['stellar_parameters']['effective_temperature'],
        Rstar=radtrans_parameters['stellar_parameters']['radius'] * petitRADTRANS.nat_cst.r_sun,
        semimajoraxis=radtrans_parameters['planetary_parameters']['orbit_semi_major_axis'],
        theta_star=radtrans_parameters['stellar_parameters']['incidence_angle']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames[
            'correlated_k_emission_cloud_calculated_radius_scattering_non-isotropic'
        ],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck_scattering.freq * 1e4,
            'spectral_radiosity': atmosphere_ck_scattering.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_correlated_k_transmission_spectrum_cloud_calculated_radius_scattering():
    mass_fractions = copy.deepcopy(radtrans_parameters['mass_fractions'])
    mass_fractions['Mg2SiO4(c)'] = \
        radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['mass_fraction']

    atmosphere_ck_scattering.calc_transm(
        temp=radtrans_parameters['temperature_isothermal'] * np.ones_like(radtrans_parameters['pressures']),
        abunds=mass_fractions,
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        Kzz=radtrans_parameters['planetary_parameters']['eddy_diffusion_coefficient'],
        fsed=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['f_sed'],
        sigma_lnorm=radtrans_parameters['cloud_parameters']['cloud_species']['Mg2SiO4(c)_cd']['sigma_log_normal']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['correlated_k_transmission_cloud_calculated_radius_scattering'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_ck_scattering.freq * 1e4,
            'transit_radius': atmosphere_ck_scattering.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )



import copy as cp
import glob

import corner
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.lines import Line2D
from scipy.stats import binned_statistic
from scipy.ndimage import uniform_filter1d
from petitRADTRANS import nat_cst as nc


def plot_specs(fig, ax, path, name, nsample, color1, color2, zorder, rebin_val=None):
    # TODO write generic plotting functions rather than copy pasting code.
    # Deprecated
    specs = sorted([f for f in glob.glob(path+'/' + name + '*.dat')])
    wlen = np.genfromtxt(specs[0])[:, 0]
    if rebin_val is not None:
        wlen = nc.running_mean(wlen, rebin_val)[::rebin_val]
    npoints = int(len(wlen))
    spectra = np.zeros((nsample, npoints))

    for i_s in range(nsample):
        if rebin_val is not None:
            spectra[i_s, :] = uniform_filter1d(np.genfromtxt(specs[i_s])[:, 1],
                                               rebin_val)[::rebin_val]
        else:
            wlen = np.genfromtxt(specs[i_s])[:, 0]

            for i_s2 in range(nsample):  # TODO check this weird loop
                spectra[i_s, :] = np.genfromtxt(specs[i_s2])[:, 1]

    sort_spec = np.sort(spectra, axis=0)
    # 3 sigma
    if int(nsample * 0.02275) > 1:
        ax.fill_between(wlen,
                        y1=sort_spec[int(nsample * 0.02275), :],
                        y2=sort_spec[int(nsample * (1. - 0.02275)), :],
                        color=color1, zorder=zorder * 2)
    # 1 sigma
    ax.fill_between(wlen,
                    y1=sort_spec[int(nsample * 0.16), :],
                    y2=sort_spec[int(nsample * 0.84), :],
                    color=color2, zorder=zorder * 2 + 1)
    return fig, ax


def plot_data(fig, ax, data, resolution=None, scaling=1.0):
    scale = data.scale_factor
    if not data.photometry:
        try:
            # Sometimes this fails, I'm not super sure why.
            resolution_data = np.mean(data.wlen[1:] / np.diff(data.wlen))
            ratio = resolution_data / resolution
            if int(ratio) > 1:
                flux, edges, _ = binned_statistic(data.wlen, data.calculate_star_radiosity, 'mean',
                                                  data.wlen.shape[0] / ratio)
                error, _, _ = np.array(binned_statistic(data.wlen, data.flux_error,
                                       'mean', data.wlen.shape[0] / ratio)) / np.sqrt(ratio)
                wlen = np.array([(edges[i] + edges[i + 1]) / 2.0 for i in range(edges.shape[0] - 1)])
            else:
                wlen = data.wlen
                error = data.flux_error
                flux = data.flux
        except:  # TODO find what is the error expected here
            wlen = data.wlen
            error = data.flux_error
            flux = data.flux
    else:
        wlen = np.mean(data.width_photometry)
        flux = data.flux
        error = data.flux_error

    marker = 'o'
    if data.photometry:
        marker = 's'
    if not data.photometry:
        ax.errorbar(wlen,
                    flux * scaling * scale,
                    yerr=error * scaling * scale,
                    marker=marker, markeredgecolor='k', linewidth=0, elinewidth=2,
                    label=data.name, zorder=10, alpha=0.9, )
    else:
        ax.errorbar(wlen,
                    flux * scaling * scale,
                    yerr=error * scaling * scale,
                    xerr=data.wlen_bins / 2., linewidth=0, elinewidth=2,
                    marker=marker, markeredgecolor='k', color='grey', zorder=10,
                    label=None, alpha=0.6)
    return fig, ax


def contour_corner(sampledict,
                   parameter_names,
                   output_file=None,
                   parameter_ranges=None,
                   parameter_plot_indices=None,
                   true_values=None,
                   short_name=None,
                   legend=False,
                   prt_plot_style=True,
                   plot_best_fit = False,
                   colors = None,
                   quintiles = [0.16, 0.5, 0.84],
                   **kwargs):
    """
    Use the corner package to plot the posterior distributions produced by pymultinest.

    Args:
        sampledict : dict
            A dictionary of samples, each sample has shape (N_Samples,N_params). The keys of the
            dictionary correspond to the names of each retrieval, and are the prefixes to the
            post_equal_weights.dat files. These are passed as arguments to retrieve.py.
            By default, this is only the current retrieval, and plots the posteriors for a single
            retrieval. If multiple names are passed, they are overplotted on the same figure.
        parameter_names : dict
            A dictionary with keys for each retrieval name, as in sampledict. Each value of the
            dictionary is the names of the parameters to beplotted, as set in the
            run_definition file.
        output_file : str
            Output file name
        parameter_ranges : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the ranges of parameters that have a range set with corner_range in the
            parameter class. Otherwise the range is +/- 4 sigma
        parameter_plot_indices : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the indices of the sample to plot, as set by the plot_in_corner
            parameter of the parameter class
        true_values : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the known values of the parameters.
        short_name : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the names to be plotted in the corner plot legend. If non, uses the
            retrieval names used as keys for sampledict
        legend : bool
            Turn the legend on or off
        prt_plot_style : bool
            Use the prt plot style, changes the colour scheme and fonts to match the rest of
            the prt plots.
        plot_best_fit : bool
            Plot the maximum likelihood values as vertical lines for each parameter.
        colors : list
            List of colours for plotting multiple retrievals.
        kwargs : dict
            Each kwarg can be one of the kwargs used in corner.corner. These can be used to adjust
            the title_kwargs,label_kwargs,hist_kwargs, hist2d_kawargs or the contour kwargs. Each
            kwarg must be a dictionary with the arguments as keys and values as the values.
    """
    if parameter_ranges is None:
        parameter_ranges = {}

    if parameter_plot_indices is None:
        parameter_plot_indices = {}

    if prt_plot_style:
        import matplotlib as mpl

        mpl.rcParams.update(mpl.rcParamsDefault)
        font = {'family': 'serif'}
        xtick = {'top': True,
                 'bottom': True,
                 'direction': 'in'}

        ytick = {'left': True,
                 'right': True,
                 'direction': 'in'}
        xmin = {'visible': True}
        ymin = {'visible': True}
        mpl.rc('xtick', **xtick)
        mpl.rc('xtick.minor', **xmin)
        mpl.rc('ytick', **ytick)
        mpl.rc('ytick.minor', **ymin)
        mpl.rc('font', **font)

        if colors is None:
            color_list = ['#009FB8', '#FF695C', '#70FF92', '#FFBB33', '#6171FF', "#FF1F69", "#52AC25", '#E574FF', "#FF261D",
                        "#B429FF"]
    else:
        if colors is None:
            color_list = [f'C{i}' for i in range(8)]  # standard matplotlib color cycle
    if colors is not None:
        color_list = colors
    handles = []
    range_list = []

    count = 0
    fig = None

    for key, samples in sampledict.items():
        if prt_plot_style and count > len(color_list):
            print("Not enough colors to continue plotting. Please add to the list.")
            print("Outputting first " + str(count) + " retrievals.")
            break

        n_samples = len(samples)
        s = n_samples

        if key not in parameter_plot_indices:
            parameter_plot_indices[key] = range(len(parameter_names[key]))
        elif parameter_plot_indices[key] is None:  # same as in the case the key doesn't exists
            parameter_plot_indices[key] = range(len(parameter_names[key]))

        if key not in parameter_ranges:
            parameter_ranges[key] = [None] * (max(parameter_plot_indices[key]) + 1)

        data_list = []
        labels_list = []

        if plot_best_fit:
            best_fit = []
            best_fit_ind = np.argmax(samples[:,-1])

            for i in parameter_plot_indices[key]:
                best_fit.append(samples[best_fit_ind][i])


        for range_i, i in enumerate(parameter_plot_indices[key]):
            data_list.append(samples[len(samples) - s:, i])
            labels_list.append(parameter_names[key][i])

            if parameter_ranges[key][i] is None:
                range_mean = np.mean(samples[len(samples) - s:, i])
                range_std = np.std(samples[len(samples) - s:, i])
                low = range_mean - 4 * range_std
                high = range_mean + 4 * range_std

                if count > 0:
                    if low > range_list[range_i][0]:
                        low = range_list[range_i][0]
                    if high < range_list[range_i][1]:
                        high = range_list[range_i][1]
                    range_take = (low, high)
                    range_list[range_i] = range_take
                else:
                    range_list.append((low, high))
            else:
                range_take = (parameter_ranges[key][i][0], parameter_ranges[key][i][1])
                if count > 0:
                    low = range_list[range_i][0]
                    high = range_list[range_i][1]
                    if range_take[0] > low:
                        low = range_take[0]
                    if range_take[1] < high:
                        high = range_take[1]
                    range_list[range_i] = (low,high)
                else:
                    range_list.append(range_take)

        if parameter_plot_indices is not None and true_values is not None:
            truths_list = []

            if plot_best_fit:
                best_fit_ind = np.argmax(samples[:, -1])

                for i in parameter_plot_indices[key]:
                    truths_list.append(samples[best_fit_ind][i])
            else:
                for i in parameter_plot_indices[key]:
                    truths_list.append(true_values[i])
        else:
            truths_list = None

        label_kwargs = None
        title_kwargs = None
        hist_kwargs = None
        hist2d_kwargs = {}
        contour_kwargs = None

        if "label_kwargs" in kwargs.keys():
            label_kwargs = kwargs["label_kwargs"]
        if "title_kwargs" in kwargs.keys():
            title_kwargs = kwargs["title_kwargs"]
        if "hist_kwargs" in kwargs.keys():
            hist_kwargs = kwargs["hist_kwargs"]
            if '#' in color_list[count]:
                value = color_list[count].lstrip('#')
                lv = len(value)
                color = tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))
                hist_kwargs["facecolor"] =  (color[0]/255,color[1]/255,color[2]/255,0.5)
            else:
                hist_kwargs["facecolor"] = color_list[count]

        if "hist2d_kwargs" in kwargs.keys():
            hist2d_kwargs = kwargs["hist2d_kwargs"]
            #hist2d_kwargs["color"] = color_list[count]
            hist2d_kwargs["fill_contours"] = True
            hist2d_kwargs["plot_density"] = True
            hist2d_kwargs["points"] = False
            hist2d_kwargs["no_fill_contours"] = False

        if "contour_kwargs" in kwargs.keys():
            contour_kwargs = kwargs["contour_kwargs"]
            contour_kwargs["colors"] = color_list[count]
        for i,label in enumerate(labels_list):
            labels_list[i] = label +'\n'
        if count == 0:
            # levels should be {1,2,3}^2/2 to enclose standard
            # confidence intervals in 2D.
            fig = corner.corner(np.array(data_list).T,
                                smooth=True,
                                title_fmt=".2f",
                                show_titles=True,
                                title_kwargs=title_kwargs,
                                labels=labels_list,
                                label_kwargs=label_kwargs,
                                range=range_list,
                                color=color_list[count],
                                quantiles=quintiles,
                                **hist2d_kwargs,
                                plot_contours=True,
                                truths=truths_list,
                                truth_color='k',
                                contour_kwargs=contour_kwargs,
                                hist_kwargs=hist_kwargs,
                                levels=[1 - np.exp(-0.5), 1 - np.exp(-1.5), 1 - np.exp(-2.5)]
                                )
        else:
            corner.corner(np.array(data_list).T,
                          fig=fig,
                          smooth=True,
                          title_fmt = None,
                          title_kwargs = None,
                          show_titles = False,
                          range = range_list,
                          color = color_list[count],
                          quantiles=quintiles,#0.16, 0.5, 0.84
                          labels = labels_list,
                          label_kwargs = label_kwargs,
                          hist2d_kwargs = hist2d_kwargs,
                          plot_contours = True,
                          contour_kwargs = contour_kwargs,
                          hist_kwargs = hist_kwargs,
                          levels=[1 - np.exp(-0.5), 1 - np.exp(-1.5), 1 - np.exp(-2.5)],
                          truths=truths_list,
                          truth_color='k'
                          )

        if short_name is None:
            label = key
        else:
            label = short_name[key]

        handles.append(Line2D([0], [0], marker='o', color=color_list[count], label=label, markersize=15))
        count += 1

    if legend:
        fig.get_axes()[2].legend(handles=handles,
                                 loc='upper right',
                                 fontsize = 20)

    if output_file is not None:
        plt.savefig(output_file, dpi=300, bbox_inches='tight')

    return fig


def contour_corner_large(sampledict,
                         parameter_names,
                         output_file=None,
                         parameter_ranges=None,
                         parameter_plot_indices=None,
                         true_values=None,
                         short_name=None,
                         legend=False,
                         prt_plot_style=True,
                         plot_best_fit=False,
                         use_labels_as_titles=True,
                         quintiles = None,
                         colors = None,
                         **kwargs):
    """
    Use the corner package to plot the posterior distributions produced by pymultinest.

    Args:
        sampledict : dict
            A dictionary of samples, each sample has shape (N_Samples,N_params). The keys of the
            dictionary correspond to the names of each retrieval, and are the prefixes to the
            post_equal_weights.dat files. These are passed as arguments to retrieve.py.
            By default, this is only the current retrieval, and plots the posteriors for a single
            retrieval. If multiple names are passed, they are overplotted on the same figure.
        parameter_names : dict
            A dictionary with keys for each retrieval name, as in sampledict. Each value of the
            dictionary is the names of the parameters to beplotted, as set in the
            run_definition file.
        output_file : str
            Output file name
        parameter_ranges : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the ranges of parameters that have a range set with corner_range in the
            parameter class. Otherwise the range is +/- 4 sigma
        parameter_plot_indices : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the indices of the sample to plot, as set by the plot_in_corner
            parameter of the parameter class
        true_values : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the known values of the parameters.
        short_name : dict
            A dictionary with keys for each retrieval name as in sampledict. Each value
            contains the names to be plotted in the corner plot legend. If non, uses the
            retrieval names used as keys for sampledict
        legend : bool
            Turn the legend on or off
        prt_plot_style : bool
            Use the prt plot style, changes the colour scheme and fonts to match the rest of
            the prt plots.
        kwargs : dict
            Each kwarg can be one of the kwargs used in corner.corner. These can be used to adjust
            the title_kwargs,label_kwargs,hist_kwargs, hist2d_kawargs or the contour kwargs. Each
            kwarg must be a dictionary with the arguments as keys and values as the values.
    """
    if parameter_ranges is None:
        parameter_ranges = {}

    if parameter_plot_indices is None:
        parameter_plot_indices = {}

    if prt_plot_style:
        import matplotlib as mpl

        mpl.rcParams.update(mpl.rcParamsDefault)
        font = {'family': 'serif'}
        xtick = {'top': True,
                 'bottom': True,
                 'direction': 'in'}

        ytick = {'left': True,
                 'right': True,
                 'direction': 'in'}
        xmin = {'visible': True}
        ymin = {'visible': True}
        mpl.rc('xtick', **xtick)
        mpl.rc('xtick.minor', **xmin)
        mpl.rc('ytick', **ytick)
        mpl.rc('ytick.minor', **ymin)
        mpl.rc('font', **font)
        
        if colors is None:
            color_list = ['#009FB8', '#FF695C', '#70FF92', '#FFBB33', '#6171FF', "#FF1F69", "#52AC25", '#E574FF', "#FF261D",
                        "#B429FF"]
    else:
        if colors is None:
            color_list = [f'C{i}' for i in range(8)]  # standard matplotlib color cycle
    if colors is not None:
        color_list = colors

    range_list = []
    handles = []
    count = 0
    fig = None

    for key, samples in sampledict.items():
        if prt_plot_style and count > len(color_list):
            print("Not enough colors to continue plotting. Please add to the list.")
            print("Outputting first " + str(count) + " retrievals.")
            break

        n_samples = len(samples)
        s = n_samples

        if key not in parameter_plot_indices:
            parameter_plot_indices[key] = range(len(parameter_names[key]))
        elif parameter_plot_indices[key] is None:  # same as in the case the key doesn't exists
            parameter_plot_indices[key] = range(len(parameter_names[key]))

        if key not in parameter_ranges:
            parameter_ranges[key] = [None] * (max(parameter_plot_indices[key]) + 1)

        data_list = []
        labels_list = []

        if use_labels_as_titles:
            titles_list = None  # if titles is None, labels are automatically used
        else:
            titles_list = []

        best_fit = None

        if plot_best_fit:
            truths_list = []
            best_fit_ind = np.argmax(samples[:,-1])
            for i in parameter_plot_indices[key]:
                truths_list.append(samples[best_fit_ind][i])

        range_list = []

        for range_i, i in enumerate(parameter_plot_indices[key]):
            data_list.append(samples[len(samples) - s:, i])
            labels_list.append(parameter_names[key][i])

            if not use_labels_as_titles:
                titles_list.append(rf"p$_{i}$")  # there is no way to make titles disappear, so use a minimal one

            if parameter_ranges[key][i] is None:
                range_mean = np.mean(samples[len(samples) - s:, i])
                range_std = np.std(samples[len(samples) - s:, i])
                low = range_mean - 4 * range_std
                high = range_mean + 4 * range_std

                if count > 0:
                    if low > range_list[range_i][0]:
                        low = range_list[range_i][0]
                    if high < range_list[range_i][1]:
                        high = range_list[range_i][1]
                    range_take = (low, high)
                    range_list[range_i] = range_take
                else:
                    range_list.append((low, high))
            else:
                range_take = (parameter_ranges[key][i][0], parameter_ranges[key][i][1])
                range_list.append(range_take)

        if parameter_plot_indices is not None and true_values is not None:
            truths_list = []

            if plot_best_fit:
                best_fit_ind = np.argmax(samples[:, -1])

                for i in parameter_plot_indices[key]:
                    truths_list.append(samples[best_fit_ind][i])
            else:
                for i in parameter_plot_indices[key]:
                    truths_list.append(true_values[key][i])
        else:
            truths_list = None

        # fig = plt.figure(figsize = (60,60),dpi=80)
        label_kwargs = None
        title_kwargs = None
        hist_kwargs = None
        hist2d_kwargs = {}
        contour_kwargs = None
        if "label_kwargs" in kwargs.keys():
            label_kwargs = kwargs["label_kwargs"]
        if "title_kwargs" in kwargs.keys():
            title_kwargs = kwargs["title_kwargs"]
        if "hist_kwargs" in kwargs.keys():
            hist_kwargs = kwargs["hist_kwargs"]
        if "hist2d_kwargs" in kwargs.keys():
            hist2d_kwargs = kwargs["hist2d_kwargs"]
        if "contour_kwargs" in kwargs.keys():
            contour_kwargs = kwargs["contour_kwargs"]
        if count == 0:
            fig = corner.corner(data=np.array(data_list).T,
                                bins=20,
                                range=range_list,
                                weights=None,
                                color=color_list[count],
                                hist_bin_factor=1,
                                smooth=True,
                                smooth1d=None,
                                labels=labels_list,
                                label_kwargs=label_kwargs,
                                titles=titles_list,
                                show_titles=True,
                                title_fmt=".2f",
                                title_kwargs=title_kwargs,
                                truths=truths_list,
                                truth_color='k',
                                scale_hist=False,
                                quantiles=quintiles,
                                verbose=False,
                                fig=None,
                                max_n_ticks=5,
                                top_ticks=False,
                                use_math_text=False,
                                reverse=False,
                                labelpad=0.0,
                                plot_contours=True,
                                contour_kwargs=contour_kwargs,
                                hist_kwargs=hist_kwargs,
                                levels=[1 - np.exp(-0.5), 1 - np.exp(-1.5), 1 - np.exp(-2.5)],
                                **hist2d_kwargs,
                                )
        else:
            corner.corner(data=np.array(data_list).T,
                          bins=20,
                          range=range_list,
                          weights=None,
                          color=color_list[count],
                          hist_bin_factor=1,
                          smooth=True,
                          smooth1d=None,
                          labels=labels_list,
                          label_kwargs=label_kwargs,
                          titles=titles_list,
                          show_titles=True,
                          title_fmt=".2f",
                          title_kwargs=title_kwargs,
                          truths=truths_list,
                          truth_color='k',
                          scale_hist=False,
                          quantiles=quintiles,
                          verbose=False,
                          fig=None,
                          max_n_ticks=5,
                          top_ticks=False,
                          use_math_text=False,
                          reverse=False,
                          labelpad=0.0,
                          plot_contours=True,
                          contour_kwargs=contour_kwargs,
                          hist_kwargs=hist_kwargs,
                          levels=[1 - np.exp(-0.5), 1 - np.exp(-1.5), 1 - np.exp(-2.5)],
                          **hist2d_kwargs,
                          )

        if short_name is None:
            label = key
        else:
            label = short_name[key]

        handles.append(Line2D([0], [0], marker='o', color=color_list[count], label=label, markersize=15))
        count += 1
    if legend:
        fig.get_axes()[2].legend(handles=handles,
                                 loc='upper right',
                                 fontsize = 20)

    if output_file is not None:
        plt.savefig(output_file, dpi=300, bbox_inches='tight')

    return fig


def nice_corner(samples,
                parameter_names,
                output_file,
                N_samples=None,
                parameter_ranges=None,
                parameter_plot_indices=None,
                true_values=None,
                max_val_ratio=None):
    """
    Paul's custom hex grid corner plots.
    Won't work with sampledict setup in retrieve.py!
    """
    font = {'family': 'serif',
            'weight': 'normal',
            'size': int(23 * 5. / len(parameter_plot_indices))}

    plt.rc('font', **font)
    plt.rc('text', usetex=True)

    if N_samples is not None:
        s = N_samples
    else:
        s = len(samples)

    try:
        if parameter_plot_indices is None:
            parameter_plot_indices = np.linspace(0, len(parameter_names) - 1,
                                                 len(parameter_names) - 1).astype('int')
    except:
        pass

    if max_val_ratio is None:
        max_val_ratio = 5.

    data_list = []
    labels_list = []
    range_list = []

    for i in parameter_plot_indices:

        data_list.append(samples[len(samples) - s:, i])
        labels_list.append(parameter_names[i])

        try:
            if parameter_ranges[i] is None:
                range_mean = np.mean(samples[len(samples) - s:, i])
                range_std = np.std(samples[len(samples) - s:, i])
                range_take = (range_mean - 4 * range_std, range_mean + 4 * range_std)
                range_list.append(range_take)
            else:
                range_list.append(parameter_ranges[i])
        except:
            range_mean = np.mean(samples[len(samples) - s:, i])
            range_std = np.std(samples[len(samples) - s:, i])
            range_take = (range_mean - 4 * range_std, range_mean + 4 * range_std)
            range_list.append(range_take)

    try:
        truths_list = []

        for i in parameter_plot_indices:
            truths_list.append(true_values[i])
    except:
        truths_list = None

    dimensions = len(parameter_plot_indices)

    # Set up the matplotlib figure
    f, axes = plt.subplots(dimensions, dimensions, figsize=(13, 13),
                           sharex='none', sharey='none')
    i_col = 0
    i_lin = 0

    try:
        ax_array = axes.flat
    except:
        ax_array = [plt.gca()]
    # print(len(axes).flat)
    for ax in ax_array:

        # print(i_col, i_lin, dimensions, len(axes.flat))
        if i_col < i_lin:
            x = cp.copy(data_list[i_col])
            y = cp.copy(data_list[i_lin])
            indexx = (x >= range_list[i_col][0]) & \
                     (x <= range_list[i_col][1])
            indexy = (y >= range_list[i_lin][0]) & \
                     (y <= range_list[i_lin][1])

            x = x[indexx & indexy]
            y = y[indexx & indexy]

            gridsize = 22
            ax.hexbin(x, y, cmap='bone_r', gridsize=gridsize,
                      vmax=int(max_val_ratio * s / gridsize ** 2.),
                      rasterized=True)

            ax.set_xlim([range_list[i_col][0], range_list[i_col][1]])
            ax.set_ylim([range_list[i_lin][0], range_list[i_lin][1]])

            try:
                ax.axhline(truths_list[i_lin], color='red',
                           linestyle='--', linewidth=2.5)
                ax.axvline(truths_list[i_col], color='red',
                           linestyle='--', linewidth=2.5)
            except:
                pass

            if i_col > 0:
                ax.get_yaxis().set_visible(False)

        elif i_col == i_lin:

            med = np.median(data_list[i_col])
            up = str(np.round(np.percentile(data_list[i_col], 84) - med, 2))
            do = str(np.round(med - np.percentile(data_list[i_col], 16), 2))
            med = str(np.round(med, 2))
            med = med.split('.')[0] + '.' + med.split('.')[1].ljust(2, '0')
            up = up.split('.')[0] + '.' + up.split('.')[1].ljust(2, '0')
            do = do.split('.')[0] + '.' + do.split('.')[1].ljust(2, '0')

            ax.set_title(med + r'$^{+' + up + '}_{-' + do + '}$',
                         fontdict=font, fontsize=int(22 * 5. / len(parameter_plot_indices)))

            use_data = cp.copy(data_list[i_col])
            index = (use_data >= range_list[i_col][0]) & \
                    (use_data <= range_list[i_col][1])
            use_data = use_data[index]
            sns.distplot(use_data, bins=22, kde=False,
                         rug=False, ax=ax, color='gray')

            ax.set_xlim([range_list[i_col][0], range_list[i_col][1]])
            ax.get_yaxis().set_visible(False)
            try:
                ax.axvline(truths_list[i_col], color='red',
                           linestyle='--', linewidth=2.5)
            except:
                pass

            ax.axvline(float(med) + float(up), color='black',
                       linestyle=':', linewidth=1.0)
            ax.axvline(float(med), color='black',
                       linestyle=':', linewidth=1.0)
            ax.axvline(float(med) - float(do), color='black',
                       linestyle=':', linewidth=1.0)

        else:
            ax.axis('off')

        i_col += 1
        if i_col % dimensions == 0:
            i_col = 0
            i_lin += 1

    for i_col in range(dimensions):
        for i_lin in range(dimensions):
            try:
                plt.sca(axes[i_lin, i_col])
            except:
                pass
            range_use = np.linspace(range_list[i_col][0],
                                    range_list[i_col][1], 5)[1:-1]

            labels_use = []
            for r in range_use:
                labels_use.append(str(np.round(r, 1)))
            plt.xticks(range_use[::2], labels_use[::2])
            plt.xlabel(labels_list[i_col])

    for i_lin in range(dimensions):
        try:
            plt.sca(axes[i_lin, 0])
        except:
            pass
        plt.ylabel(labels_list[i_lin])

    plt.subplots_adjust(wspace=0, hspace=0)
    if dimensions == 1:
        plt.tight_layout(h_pad=0, w_pad=0)
    plt.savefig(output_file)
    # plt.show()
    # plt.clf()



from __future__ import print_function
import astropy.constants as anc
import numpy as np
import scipy.constants as snc

import os as os
import sys

# Natural constants
# Everything is in cgs!
# Defined constants
c = snc.c * 1e2
h = snc.h * 1e7
kB = snc.k * 1e7
nA = snc.N_A
e = snc.e * np.sqrt(1e9 / (4 * snc.pi * snc.epsilon_0))

# Measured constants
G = snc.G * 1e3
m_elec = snc.m_e * 1e3

# Derived exact constants
sigma = snc.sigma * 1e3
L0 = snc.physical_constants['Loschmidt constant (273.15 K, 101.325 kPa)'][0] * 1e-6
R = snc.R

# Units definitions
bar = 1e6
atm = snc.atm * 1e1
AU = snc.au * 1e2
pc = snc.parsec * 1e2
light_year = snc.light_year * 1e2
amu = snc.physical_constants['atomic mass constant'][0] * 1e3

# Astronomical constants
r_sun = anc.R_sun.cgs.value
r_jup = anc.R_jup.cgs.value
r_earth = anc.R_earth.cgs.value
m_sun = anc.M_sun.cgs.value
m_jup = anc.M_jup.cgs.value
m_earth = anc.M_earth.cgs.value
l_sun = anc.L_sun.cgs.value

r_jup_mean = 6.9911e9
s_earth = 1.3654e6  # erg.s-1.cm-2, source: https://agupubs.onlinelibrary.wiley.com/doi/full/10.1029/2010GL045777

# Molecular weights in amu
molecular_weight = {
    'H2O': 18.,
    'O2': 32.,
    'N2': 28.,
    'CH4': 16.,
    'CO2': 44.,
    'CO': 28.,
    'H2': 2.,
    'He': 4.
}

def b(T,nu):
    ''' Returns the Planck function :math:`B_{\\nu}(T)` in units of
    :math:`\\rm erg/s/cm^2/Hz/steradian`.

    Args:
        T (float):
            Temperature in K.
        nu:
            numpy array containing the frequency in Hz.
    '''

    retVal = 2.*h*nu**3./c**2.
    retVal = retVal / (np.exp(h*nu/kB/T)-1.)
    return retVal

pathinp = os.environ.get("pRT_input_data_path")
if pathinp == None:
    raise OSError(f"Path to input data not specified!\n"
                  f"Please set pRT_input_data_path variable in .bashrc / .bash_profile or specify path via\n"
                  f">>> import os"
                  f">>> os.environ['pRT_input_data_path'] = 'absolute/path/of/the/folder/input_data'\n"
                  f"before creating a Radtrans object or loading the nat_cst module.\n"
                  f"(this will become unnecessary in a future update)"
                  )
spec_path = pathinp + '/stellar_specs'

description = np.genfromtxt(spec_path+'/stellar_params.dat')
logTempGrid = description[:,0]
StarRadGrid = description[:,1]

specDats = []
for i in range(len(logTempGrid)):
    specDats.append(np.genfromtxt(spec_path+'/spec_'+ \
                    str(int(i)).zfill(2)+'.dat'))

def get_PHOENIX_spec(temperature):
    ''' Returns a matrix where the first column is the wavelength in cm
    and the second is the stellar flux :math:`F_\\nu` in units of
    :math:`\\rm erg/cm^2/s/Hz`, at the surface of the star.
    The spectra are PHOENIX models from (Husser et al. 2013), the spectral
    grid used here was described in van Boekel et al. (2012).

    Args:
        temperature (float):
            stellar effective temperature in K.
    '''
    logTemp = np.log10(temperature)
    interpolationIndex = np.searchsorted(logTempGrid, logTemp)

    if interpolationIndex == 0:

        specDat = specDats[0]
        print('Warning, input temperature is lower than minimum grid temperature.')
        print('Taking F = F_grid(minimum grid temperature), normalized to desired')
        print('input temperature.')

    elif interpolationIndex == len(logTempGrid):

        specDat = specDats[int(len(logTempGrid)-1)]
        print('Warning, input temperature is higher than maximum grid temperature.')
        print('Taking F = F_grid(maximum grid temperature), normalized to desired')
        print('input temperature.')

    else:

        weightHigh = (logTemp-logTempGrid[interpolationIndex-1]) / \
          (logTempGrid[interpolationIndex]-logTempGrid[interpolationIndex-1])

        weightLow = 1. - weightHigh

        specDatLow = specDats[int(interpolationIndex-1)]

        specDatHigh = specDats[int(interpolationIndex)]

        specDat = np.zeros_like(specDatLow)

        specDat[:,0] = specDatLow[:,0]
        specDat[:,1] = weightLow * specDatLow[:,1] + \
          weightHigh * specDatHigh[:,1]

    freq = c/specDat[:,0]
    flux = specDat[:,1]
    norm = -np.sum((flux[1:]+flux[:-1])*np.diff(freq))/2.

    specDat[:,1] = flux/norm*sigma*temperature**4.

    return specDat

def get_PHOENIX_spec_rad(temperature):
    '''
    Returns a matrix where the first column is the wavelength in cm
    and the second is the stellar flux :math:`F_\\nu` in units of
    :math:`\\rm erg/cm^2/s/Hz`, at the surface of the star.
    The spectra are PHOENIX models from (Husser et al. 2013), the spectral
    grid used here was described in van Boekel et al. (2012).

    UPDATE: It also returns a float that is the corresponding estimate
    of the stellar radius.

    Args:
        temperature (float):
            stellar effective temperature in K.
    '''

    logTemp = np.log10(temperature)
    interpolationIndex = np.searchsorted(logTempGrid, logTemp)

    if interpolationIndex == 0:

        specDat = specDats[0]
        radius = StarRadGrid[0]
        print('Warning, input temperature is lower than minimum grid temperature.')
        print('Taking F = F_grid(minimum grid temperature), normalized to desired')
        print('input temperature.')

    elif interpolationIndex == len(logTempGrid):

        specDat = specDats[int(len(logTempGrid)-1)]
        radius = StarRadGrid[int(len(logTempGrid)-1)]
        print('Warning, input temperature is higher than maximum grid temperature.')
        print('Taking F = F_grid(maximum grid temperature), normalized to desired')
        print('input temperature.')

    else:

        weightHigh = (logTemp-logTempGrid[interpolationIndex-1]) / \
          (logTempGrid[interpolationIndex]-logTempGrid[interpolationIndex-1])

        weightLow = 1. - weightHigh

        specDatLow = specDats[int(interpolationIndex-1)]

        specDatHigh = specDats[int(interpolationIndex)]

        specDat = np.zeros_like(specDatLow)

        specDat[:,0] = specDatLow[:,0]
        specDat[:,1] = weightLow * specDatLow[:,1] + \
          weightHigh * specDatHigh[:,1]
        radius = weightLow * StarRadGrid[int(interpolationIndex-1)] + \
          weightHigh * StarRadGrid[int(interpolationIndex)]

    freq = c/specDat[:,0]
    flux = specDat[:,1]
    norm = -np.sum((flux[1:]+flux[:-1])*np.diff(freq))/2.

    specDat[:,1] = flux/norm*sigma*temperature**4.

    return specDat,radius

# Future functions
def convolve_rebin(input_wavelengths, input_flux,
                   instrument_resolving_power, pixel_sampling, instrument_wavelength_range):
    from scipy.ndimage import gaussian_filter1d
    import petitRADTRANS.fort_rebin as fr
    """
    Function to convolve observation with instrument obs and rebin to pixels of detector.
    Create mock observation for high-res spectrograph.

    Args:
        input_wavelengths: (cm) wavelengths of the input spectrum
        input_flux: flux of the input spectrum
        instrument_resolving_power: resolving power of the instrument
        pixel_sampling: number of pixels per resolution elements (i.e. how many px in one LSF FWHM, usually 2)
        instrument_wavelength_range: (um) wavelength range of the instrument

    Returns:
        flux_lsf: flux altered by the instrument's LSF
        freq_out: (Hz) frequencies of the rebinned flux, in descending order
        flux_rebin: the rebinned flux
    """
    # From talking to Ignas: delta lambda of resolution element is the FWHM of the instrument's LSF (here: a gaussian)
    sigma_lsf = 1. / instrument_resolving_power / (2. * np.sqrt(2. * np.log(2.)))

    # The input resolution of petitCODE is 1e6, but just compute
    # it to be sure, or more versatile in the future.
    # Also, we have a log-spaced grid, so the resolution is constant
    # as a function of wavelength
    model_resolving_power = np.mean(
        (input_wavelengths[1:] + input_wavelengths[:-1]) / (2. * np.diff(input_wavelengths))
    )

    # Calculate the sigma to be used in the gauss filter in units of input frequency bins
    sigma_lsf_gauss_filter = sigma_lsf * model_resolving_power

    flux_lsf = gaussian_filter1d(
        input=input_flux,
        sigma=sigma_lsf_gauss_filter,
        mode='reflect'
    )

    if np.size(instrument_wavelength_range) == 2:  # TODO check if this is still working
        wavelength_out_borders = np.logspace(
            np.log10(instrument_wavelength_range[0]),
            np.log10(instrument_wavelength_range[1]),
            int(pixel_sampling * instrument_resolving_power
                * np.log(instrument_wavelength_range[1] / instrument_wavelength_range[0]))
        )
        wavelengths_out = (wavelength_out_borders[1:] + wavelength_out_borders[:-1]) / 2.
    elif np.size(instrument_wavelength_range) > 2:
        wavelengths_out = instrument_wavelength_range
    else:
        raise ValueError(f"instrument wavelength must be of size 2 or more, "
                         f"but is of size {np.size(instrument_wavelength_range)}: {instrument_wavelength_range}")

    flux_rebin = fr.rebin_spectrum(input_wavelengths, flux_lsf, wavelengths_out)

    return flux_lsf, wavelengths_out, flux_rebin


def radiosity_erg_hz2radiosity_erg_cm(radiosity_erg_hz, frequency):
    """Convert a radiosity from erg.s-1.cm-2.sr-1/Hz to erg.s-1.cm-2.sr-1/cm at a given frequency.  # TODO move to physics

    Steps:
        [cm] = c[cm.s-1] / [Hz]
        => d[cm]/d[Hz] = d(c / [Hz])/d[Hz]
        => d[cm]/d[Hz] = c / [Hz]**2
        => d[Hz]/d[cm] = [Hz]**2 / c
        integral of flux must be conserved: radiosity_erg_cm * d[cm] = radiosity_erg_hz * d[Hz]
        radiosity_erg_cm = radiosity_erg_hz * d[Hz]/d[cm]
        => radiosity_erg_cm = radiosity_erg_hz * frequency**2 / c

    Args:
        radiosity_erg_hz: (erg.s-1.cm-2.sr-1/Hz)
        frequency: (Hz)

    Returns:
        (erg.s-1.cm-2.sr-1/cm) the radiosity in converted units
    """
    return radiosity_erg_hz * frequency ** 2 / c



"""Stores functions that convert files from a format to another.

The functions in this module are stored for the sake of keeping trace of changes made to files. They are intended to be
used only once.
"""
import os

import h5py
import numpy as np

import petitRADTRANS.nat_cst as nc


def __phoenix_spec_dat2h5():
    """
    Convert a PHOENIX stellar spectrum in .dat format to HDF5 format.
    """
    # Load the stellar parameters
    description = np.genfromtxt(nc.spec_path + os.path.sep + 'stellar_params.dat')

    # Initialize the grids
    log_temp_grid = description[:, 0]
    star_rad_grid = description[:, 1]

    # Load the corresponding numbered spectral files
    spec_dats = []

    for spec_num in range(len(log_temp_grid)):
        spec_dats.append(np.genfromtxt(nc.spec_path + '/spec_'
                                       + str(int(spec_num)).zfill(2) + '.dat'))

    # Write the HDF5 file
    with h5py.File("stellar_spectra.h5", "w") as f:
        t_eff = f.create_dataset(
            name='log10_effective_temperature',
            data=log_temp_grid
        )
        t_eff.attrs['units'] = 'log10(K)'

        radius = f.create_dataset(
            name='radius',
            data=star_rad_grid
        )
        radius.attrs['units'] = 'R_sun'

        mass = f.create_dataset(
            name='mass',
            data=description[:, 2]
        )
        mass.attrs['units'] = 'M_sun'

        spectral_type = f.create_dataset(
            name='spectral_type',
            data=description[:, -1]
        )
        spectral_type.attrs['units'] = 'None'

        wavelength = f.create_dataset(
            name='wavelength',
            data=np.asarray(spec_dats)[0, :, 0]
        )
        wavelength.attrs['units'] = 'cm'

        spectral_radiosity = f.create_dataset(
            name='spectral_radiosity',
            data=np.asarray(spec_dats)[:, :, 1]
        )
        spectral_radiosity.attrs['units'] = 'erg/s/cm^2/Hz'



"""Test the petitRADTRANS retrieval module.

Essentially go through a simplified version of the tutorial, and compare the results with previous ones.
C.f. (https://petitradtrans.readthedocs.io/en/latest/content/notebooks/getting_started.html).
"""
import json
import os

os.environ["OMP_NUM_THREADS"] = "1"  # to not have numpy start parallelizing on its own
import numpy as np

from petitRADTRANS.retrieval.util import gaussian_prior
from petitRADTRANS.retrieval.util import calc_MMW

from .context import petitRADTRANS
from .utils import tests_results_directory, reference_filenames, radtrans_parameters


relative_tolerance = 1e-6  # relative tolerance when comparing with older results
max_number_of_tests = 3


def init_run():
    # Since our retrieval has already run before, we'll set the mode to 'evaluate' so we can make some plots.
    run_definition_simple = petitRADTRANS.retrieval.RetrievalConfig(
        retrieval_name="test",
        run_mode="retrieval",  # This must be 'retrieval' to run PyMultiNest
        AMR=False,  # We won't be using adaptive mesh refinement for the pressure grid
        pressures=radtrans_parameters['pressures'],
        scattering=False  # This would turn on scattering when calculating emission spectra
    )
    # Scattering is automatically included for transmission spectra

    # Fixed parameters
    run_definition_simple.add_parameter(
        'Rstar',
        False,
        value=radtrans_parameters['stellar_parameters']['radius'] * petitRADTRANS.nat_cst.r_sun
    )

    # Log of the surface gravity
    run_definition_simple.add_parameter(
        'log_g',
        False,
        value=np.log10(radtrans_parameters['planetary_parameters']['surface_gravity'])
    )

    # Retrieved parameters
    # Prior functions
    def prior_planet_radius(x):
        return petitRADTRANS.retrieval.util.uniform_prior(
            cube=x,
            x1=radtrans_parameters['retrieval_parameters']['planetary_radius_bounds'][0]
            * petitRADTRANS.nat_cst.r_jup_mean,
            x2=radtrans_parameters['retrieval_parameters']['planetary_radius_bounds'][1]
            * petitRADTRANS.nat_cst.r_jup_mean,
        )

    def prior_temperature(x):
        return petitRADTRANS.retrieval.util.uniform_prior(
            cube=x,
            x1=radtrans_parameters['retrieval_parameters']['intrinsic_temperature_bounds'][0],
            x2=radtrans_parameters['retrieval_parameters']['intrinsic_temperature_bounds'][1]
        )

    def prior_cloud_pressure(x):
        return petitRADTRANS.retrieval.util.uniform_prior(
            cube=x,
            x1=radtrans_parameters['retrieval_parameters']['log10_cloud_pressure_bounds'][0],
            x2=radtrans_parameters['retrieval_parameters']['log10_cloud_pressure_bounds'][1]
        )

    # Planet radius
    run_definition_simple.add_parameter(
        'R_pl',
        True,
        transform_prior_cube_coordinate=prior_planet_radius
    )

    # Intrinsic temperature
    run_definition_simple.add_parameter(
        'Temperature',
        True,
        transform_prior_cube_coordinate=prior_temperature
    )

    # Include a grey cloud as well to see what happens
    run_definition_simple.add_parameter(
        "log_Pcloud",
        True,
        transform_prior_cube_coordinate=prior_cloud_pressure
    )

    # Spectrum parameters
    run_definition_simple.set_rayleigh_species(radtrans_parameters['spectrum_parameters']['rayleigh_species'])
    run_definition_simple.set_continuum_opacities(radtrans_parameters['spectrum_parameters']['continuum_opacities'])

    run_definition_simple.set_line_species(
        radtrans_parameters['spectrum_parameters']['line_species_correlated_k'],
        eq=False,
        abund_lim=(
            radtrans_parameters['retrieval_parameters']['log10_species_mass_fractions_bounds'][0],
            radtrans_parameters['retrieval_parameters']['log10_species_mass_fractions_bounds'][1]
        )  # prior: min = abund_lim[0], max = min + abund_lim[1]
    )

    # Load data
    run_definition_simple.add_data(
        name='test',
        path=reference_filenames['mock_observation_transmission'],
        model_generating_function=retrieval_model_spec_iso,
        opacity_mode='c-k',
        data_resolution=radtrans_parameters['mock_observation_parameters']['resolving_power'],
        model_resolution=radtrans_parameters['mock_observation_parameters']['resolving_power'] * 2
    )

    # Plot parameters
    # Corner plot
    run_definition_simple.parameters['R_pl'].plot_in_corner = True
    run_definition_simple.parameters['R_pl'].corner_label = r'$R_{\rm P}$ ($\rm R_{Jup}$)'
    run_definition_simple.parameters['R_pl'].corner_transform = lambda x: x / petitRADTRANS.nat_cst.r_jup_mean
    run_definition_simple.parameters['Temperature'].plot_in_corner = True
    run_definition_simple.parameters['Temperature'].corner_label = "Temp"
    run_definition_simple.parameters['log_Pcloud'].plot_in_corner = True
    run_definition_simple.parameters['log_Pcloud'].corner_label = r"log P$_{\rm cloud}$"
    run_definition_simple.parameters['log_Pcloud'].corner_ranges = [-6, 2]

    for spec in run_definition_simple.line_species:
        run_definition_simple.parameters[spec].plot_in_corner = True
        run_definition_simple.parameters[spec].corner_ranges = [-6.0, 0.0]

    # Spectrum plot
    run_definition_simple.plot_kwargs["spec_xlabel"] = 'Wavelength [micron]'
    run_definition_simple.plot_kwargs["spec_ylabel"] = r'$(R_{\rm P}/R_*)^2$ [ppm]'
    run_definition_simple.plot_kwargs["y_axis_scaling"] = 1e6  # so we have units of ppm
    run_definition_simple.plot_kwargs["xscale"] = 'linear'
    run_definition_simple.plot_kwargs["yscale"] = 'linear'

    run_definition_simple.plot_kwargs["nsample"] = 10

    # Temperature profile plot
    run_definition_simple.plot_kwargs["take_PTs_from"] = 'test'
    run_definition_simple.plot_kwargs["temp_limits"] = [150, 3000]
    run_definition_simple.plot_kwargs["press_limits"] = [1e2, 1e-5]

    return run_definition_simple


# Atmospheric model
def retrieval_model_spec_iso(prt_object, parameters, pt_plot_mode=None, AMR=False):
    """
    This model computes a transmission spectrum based on free retrieval chemistry
    and an isothermal temperature-pressure profile.

    Args:
        prt_object : object
            An instance of the pRT class, with optical properties as defined in the RunDefinition.
        parameters : dict
            Dictionary of required parameters:
                Rstar : Radius of the host star [cm]
                log_g : Log of surface gravity
                R_pl : planet radius [cm]
                Temperature : Isothermal temperature [K]
                species : Log abundances for each species used in the retrieval
                log_Pcloud : optional, cloud base pressure of a grey cloud deck.
        pt_plot_mode:
            Return only the pressure-temperature profile for plotting. Evaluate mode only. Mandatory.
        AMR:
            Adaptive mesh refinement. Use the high resolution pressure grid around the cloud base. Mandatory.

    Returns:
        wavelength_model : np.array
            Wavlength array of computed model, not binned to data [um]
        spectrum_model : np.array
            Computed transmission spectrum R_pl**2/Rstar**2
    """
    # Make the P-T profile
    pressures = prt_object.press * 1e-6  # bar to cgs
    temperatures = parameters['Temperature'].value * np.ones_like(pressures)

    # Make the abundance profiles
    abundances = {}
    m_sum = 0.0  # Check that the total mass fraction of all species is <1

    for species in prt_object.line_species:
        spec = species.split('_R_')[0]  # deal with the naming scheme for binned down opacities (see below)
        abundances[species] = 10 ** parameters[spec].value * np.ones_like(pressures)
        m_sum += 10 ** parameters[spec].value

    abundances['H2'] = radtrans_parameters['mass_fractions']['H2'] * (1.0 - m_sum) * np.ones_like(pressures)
    abundances['He'] = radtrans_parameters['mass_fractions']['He'] * (1.0 - m_sum) * np.ones_like(pressures)

    # Find the mean molecular weight in each layer
    mmw = calc_MMW(abundances)

    # Calculate the spectrum
    prt_object.calc_transm(
        temperatures,
        abundances,
        10 ** parameters['log_g'].value,
        mmw,
        R_pl=parameters['R_pl'].value,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure'],
        Pcloud=10 ** parameters['log_Pcloud'].value
    )

    # Transform the outputs into the units of our data.
    wlen_model = petitRADTRANS.nat_cst.c / prt_object.freq * 1e4  # wlen in micron
    spectrum_model = (prt_object.transm_rad / parameters['Rstar'].value) ** 2.

    return wlen_model, spectrum_model


run_definition = init_run()


def test_list_available_species():
    run_definition.list_available_line_species()


def test_simple_retrieval(test_number=0, max_test_number=max_number_of_tests):
    retrieval = petitRADTRANS.retrieval.Retrieval(
        run_definition,
        output_dir=tests_results_directory,
        sample_spec=radtrans_parameters['retrieval_parameters']['sample_spectrum_output'],
        ultranest=radtrans_parameters['retrieval_parameters']['ultranest']  # if False, use PyMultiNest
    )

    try:
        retrieval.run(
            sampling_efficiency=radtrans_parameters['retrieval_parameters']['sampling_efficiency'],
            n_live_points=radtrans_parameters['retrieval_parameters']['n_live_points'],
            const_efficiency_mode=radtrans_parameters['retrieval_parameters']['const_efficiency_mode'],
            resume=radtrans_parameters['retrieval_parameters']['resume']
        )

        # Just check if get_samples works
        sample_dict, parameter_dict = retrieval.get_samples()

        # Get results and reference
        with open(reference_filenames['pymultinest_parameter_analysis']) as f:
            reference = json.load(f)

        new_result_file = os.path.join(
            tests_results_directory,
            'out_PMN',
            os.path.basename(reference_filenames['pymultinest_parameter_analysis'])
        )

        with open(new_result_file) as f:
            new_results = json.load(f)

        # Check if retrieved parameters are in +/- 1 sigma of the previous retrieved parameters
        for i, marginal in enumerate(new_results['marginals']):
            assert marginal['median'] - reference['marginals'][i]['sigma'] \
                   <= reference['marginals'][i]['median'] \
                   <= marginal['median'] + reference['marginals'][i]['sigma']
    except AssertionError as error_message:
        test_number += 1

        if test_number < max_test_number:
            test_simple_retrieval(test_number=test_number, max_test_number=max_test_number)
        else:
            raise AssertionError(
                f"Retrievals are expected to give results within a +/- 1 sigma uncertainty range with a probability of "
                f"~68% for each parameters."
                f"To take that into account, {test_number} tests were performed, "
                f"but all failed to fall within this range "
                f"compared to the results of the previous version.\n"
                f"Complete error message was: \n" +
                str(error_message)
            )



"""Test petitRADTRANS utility functions.
"""
import numpy as np

from .context import petitRADTRANS
from .utils import reference_filenames, radtrans_parameters, temperature_guillot_2010

relative_tolerance = 1e-6


def test_guillot_2010_temperature_profile():
    # Load expected results
    reference_data = np.load(reference_filenames['guillot_2010'])
    print(f"Comparing generated spectrum to result from petitRADTRANS-{reference_data['prt_version']}...")
    temperature_ref = reference_data['temperature']
    pressure_ref = reference_data['pressure']

    # Check if temperature is as expected
    assert np.allclose(
        radtrans_parameters['pressures'],
        pressure_ref,
        rtol=relative_tolerance,
        atol=0
    )

    assert np.allclose(
        temperature_guillot_2010,
        temperature_ref,
        rtol=relative_tolerance,
        atol=0
    )


def test_planck_function():
    frequencies = petitRADTRANS.nat_cst.c / np.linspace(
        radtrans_parameters['spectrum_parameters']['wavelength_range_correlated_k'][0] * 1e-4,
        radtrans_parameters['spectrum_parameters']['wavelength_range_correlated_k'][1] * 1e-4,
        10
    )

    planck = petitRADTRANS.nat_cst.b(
        radtrans_parameters['stellar_parameters']['effective_temperature'],
        frequencies
    )

    reference_spectral_radiance = np.array([
        3.65580443e-05, 3.64376693e-05, 3.62173874e-05, 3.59133996e-05, 3.55399032e-05,
        3.51092680e-05, 3.46322194e-05, 3.41180180e-05, 3.35746299e-05, 3.30088840e-05
    ])  # (erg.s-1.sr-1.cm-2.Hz-1)

    assert np.allclose(
        planck,
        reference_spectral_radiance,
        rtol=relative_tolerance,
        atol=0
    )


def test_stellar_model():
    stellar_spectrum = petitRADTRANS.nat_cst.get_PHOENIX_spec(
        radtrans_parameters['stellar_parameters']['effective_temperature']
    )

    reference_spectral_radiosity = np.array([
        7.66275355e-114, 1.93595034e-058, 9.90294639e-023, 2.70611283e-008, 8.19891143e-005,
        1.05038448e-004, 2.47953373e-005, 3.87505010e-006, 5.57445915e-007, 7.72638833e-008
    ])  # (erg.s-1.cm-2.Hz-1)

    reference_wavenumber = np.array([
        1.0002808e-06, 2.6753539e-06, 7.1555055e-06, 1.9138131e-05, 5.1186878e-05,
        1.3690452e-04, 3.6616507e-04, 9.7934564e-04, 2.6193606e-03, 7.0057423e-03
    ])  # (cm)

    assert np.allclose(
        stellar_spectrum[1::3500, 0],
        reference_wavenumber,
        rtol=relative_tolerance,
        atol=0
    )

    assert np.allclose(
        stellar_spectrum[1::3500, 1],
        reference_spectral_radiosity,
        rtol=relative_tolerance,
        atol=0
    )



"""Test the radtrans module in line-by-line mode.

Essentially go through a simplified version of the tutorial, and compare the results with previous ones.
C.f. (https://petitradtrans.readthedocs.io/en/latest/content/notebooks/getting_started.html).

Do not change the parameters used to generate the comparison files, including input_data files, when running the tests.
"""
import copy

import numpy as np

from .context import petitRADTRANS
from .utils import compare_from_reference_file, \
    reference_filenames, radtrans_parameters, temperature_guillot_2010, temperature_isothermal

relative_tolerance = 1e-6  # relative tolerance when comparing with older results


# Initializations
def init_radtrans_line_by_line():
    atmosphere = petitRADTRANS.radtrans.Radtrans(
        line_species=radtrans_parameters['spectrum_parameters']['line_species_line_by_line'],
        rayleigh_species=radtrans_parameters['spectrum_parameters']['rayleigh_species'],
        continuum_opacities=radtrans_parameters['spectrum_parameters']['continuum_opacities'],
        wlen_bords_micron=radtrans_parameters['spectrum_parameters']['wavelength_range_line_by_line'],
        mode='lbl'
    )

    atmosphere.setup_opa_structure(radtrans_parameters['pressures'])

    return atmosphere


def init_spectral_model_line_by_line():
    spectral_model = petitRADTRANS.containers.spectral_model.SpectralModel(
        # Radtrans object parameters
        pressures=radtrans_parameters['pressures'],  # bar
        line_species=radtrans_parameters['spectrum_parameters']['line_species_line_by_line'],
        rayleigh_species=radtrans_parameters['spectrum_parameters']['rayleigh_species'],
        continuum_opacities=radtrans_parameters['spectrum_parameters']['continuum_opacities'],
        do_scat_emis=False,  # is False by default on Radtrans but True by default on SpectralModel
        opacity_mode='lbl',
        # Temperature profile parameters: generate a Guillot temperature profile
        temperature_profile_mode='guillot',
        temperature=radtrans_parameters['temperature_guillot_2010_parameters']['equilibrium_temperature'],  # K
        intrinsic_temperature=radtrans_parameters['temperature_guillot_2010_parameters']['intrinsic_temperature'],
        metallicity=1.0,
        guillot_temperature_profile_gamma=radtrans_parameters['temperature_guillot_2010_parameters']['gamma'],
        guillot_temperature_profile_kappa_ir_z0=radtrans_parameters['temperature_guillot_2010_parameters'][
            'infrared_mean_opacity'],
        # Chemical parameters
        use_equilibrium_chemistry=False,
        imposed_mass_mixing_ratios=radtrans_parameters['mass_fractions'],
        # Transmission spectrum parameters (radtrans.calc_transm)
        planet_radius=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,  # cm
        planet_surface_gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        reference_pressure=radtrans_parameters['planetary_parameters']['reference_pressure'],  # bar
        # cloud_pressure=1e-1,
        # Instrument parameters
        new_resolving_power=radtrans_parameters['mock_observation_parameters']['high_resolution_resolving_power'],
        output_wavelengths=np.arange(
            radtrans_parameters['mock_observation_parameters']['wavelength_range_high_resolution'][0],
            radtrans_parameters['mock_observation_parameters']['wavelength_range_high_resolution'][1],
            radtrans_parameters['mock_observation_parameters']['wavelength_range_high_resolution'][0] /
            radtrans_parameters['mock_observation_parameters']['high_resolution_resolving_power'] / 2
        ),  # um
        wavelengths_boundaries=radtrans_parameters['spectrum_parameters']['wavelength_range_line_by_line'],
        # Scaling parameters
        star_radius=radtrans_parameters['stellar_parameters']['radius'] * petitRADTRANS.nat_cst.r_sun,  # cm
        # Orbital parameters
        star_mass=radtrans_parameters['stellar_parameters']['radius'] * petitRADTRANS.nat_cst.m_sun,  # g
        semi_major_axis=radtrans_parameters['planetary_parameters']['orbit_semi_major_axis'],  # cm
        orbital_phases=np.linspace(
            radtrans_parameters['mock_observation_parameters']['orbital_phase_range'][0],
            radtrans_parameters['mock_observation_parameters']['orbital_phase_range'][1],
            radtrans_parameters['mock_observation_parameters']['n_exposures']
        ),
        system_observer_radial_velocities=np.linspace(
            radtrans_parameters['mock_observation_parameters']['system_observer_radial_velocities_range'][0],
            radtrans_parameters['mock_observation_parameters']['system_observer_radial_velocities_range'][1],
            radtrans_parameters['mock_observation_parameters']['n_exposures']
        ),  # cm.s-1
        planet_rest_frame_velocity_shift=radtrans_parameters['mock_observation_parameters'][
            "planet_rest_frame_velocity_shift"],  # cm.s-1
        # Reprocessing parameters
        # uncertainties=model_uncertainties,
        # airmass=airmasses,
        # tellurics_mask_threshold=tellurics_mask_threshold,
        # polynomial_fit_degree=2,
        # apply_throughput_removal=True,
        # apply_telluric_lines_removal=True,
        # Special parameters
        # Test addition of a useless parameter
        yet_another_useless_parameter42="irrelevant string"
    )

    def calculate_mean_molar_masses(pressures, **kwargs):
        return radtrans_parameters['mean_molar_mass'] * np.ones(pressures.size)

    # Test custom function
    spectral_model.calculate_mass_mixing_ratios = \
        petitRADTRANS.containers.spectral_model.BaseSpectralModel.calculate_mass_mixing_ratios
    spectral_model.calculate_mean_molar_masses = \
        calculate_mean_molar_masses

    radtrans = spectral_model.get_radtrans()

    return spectral_model, radtrans


atmosphere_lbl = init_radtrans_line_by_line()
spectral_model_lbl, radtrans_spectral_model_lbl = init_spectral_model_line_by_line()


def test_line_by_line_emission_spectrum():
    # Calculate an emission spectrum
    atmosphere_lbl.calc_flux(
        temp=temperature_guillot_2010,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['line_by_line_emission'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_lbl.freq * 1e4,
            'spectral_radiosity': atmosphere_lbl.flux
        },
        relative_tolerance=relative_tolerance
    )


def test_line_by_line_transmission_spectrum():
    # Calculate a transmission spectrum
    atmosphere_lbl.calc_transm(
        temp=temperature_isothermal,
        abunds=radtrans_parameters['mass_fractions'],
        gravity=radtrans_parameters['planetary_parameters']['surface_gravity'],
        mmw=radtrans_parameters['mean_molar_mass'],
        R_pl=radtrans_parameters['planetary_parameters']['radius'] * petitRADTRANS.nat_cst.r_jup_mean,
        P0_bar=radtrans_parameters['planetary_parameters']['reference_pressure']
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['line_by_line_transmission'],
        comparison_dict={
            'wavelength': petitRADTRANS.nat_cst.c / atmosphere_lbl.freq * 1e4,
            'transit_radius': atmosphere_lbl.transm_rad / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def test_line_by_line_spectral_model_emission():
    spectral_model = copy.deepcopy(spectral_model_lbl)
    spectral_model.model_parameters['is_orbiting'] = False

    wavelengths, spectral_radiosities = spectral_model.get_spectrum_model(
        radtrans=radtrans_spectral_model_lbl,
        mode='emission',
        parameters=None,
        update_parameters=True,
        telluric_transmittances=None,
        instrumental_deformations=None,
        noise_matrix=None,
        scale=False,
        shift=False,
        convolve=False,
        rebin=False,
        reduce=False
    )

    assert np.allclose(
        spectral_model.temperatures,
        temperature_guillot_2010,
        atol=0,
        rtol=relative_tolerance
    )

    for species in spectral_model.line_species:
        assert np.allclose(
            spectral_model.mass_mixing_ratios[species],
            radtrans_parameters['mass_fractions'][species],
            atol=0,
            rtol=relative_tolerance
        )

    assert np.allclose(
        spectral_model.mean_molar_masses,
        radtrans_parameters['mean_molar_mass'],
        atol=0,
        rtol=relative_tolerance
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['line_by_line_emission'],
        comparison_dict={
            'wavelength': wavelengths,
            'spectral_radiosity': petitRADTRANS.physics.radiosity_erg_cm2radiosity_erg_hz(
                radiosity_erg_cm=spectral_radiosities * 1e7,  # W to erg
                wavelength=wavelengths * 1e-4  # um to cm
            )
        },
        relative_tolerance=relative_tolerance
    )


def test_line_by_line_spectral_model_transmission():
    spectral_model = copy.deepcopy(spectral_model_lbl)

    spectral_model.model_parameters['temperature_profile_mode'] = 'isothermal'
    spectral_model.model_parameters['temperature'] = temperature_isothermal

    wavelengths, transit_radii = spectral_model.get_spectrum_model(
        radtrans=radtrans_spectral_model_lbl,
        mode='transmission',
        parameters=None,
        update_parameters=True,
        telluric_transmittances=None,
        instrumental_deformations=None,
        noise_matrix=None,
        scale=False,
        shift=False,
        convolve=False,
        rebin=False,
        reduce=False
    )

    assert np.allclose(
        spectral_model.temperatures,
        temperature_isothermal,
        atol=0,
        rtol=relative_tolerance
    )

    for species in spectral_model.line_species:
        assert np.allclose(
            spectral_model.mass_mixing_ratios[species],
            radtrans_parameters['mass_fractions'][species],
            atol=0,
            rtol=relative_tolerance
        )

    assert np.allclose(
        spectral_model.mean_molar_masses,
        radtrans_parameters['mean_molar_mass'],
        atol=0,
        rtol=relative_tolerance
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['line_by_line_transmission'],
        comparison_dict={
            'wavelength': wavelengths,
            'transit_radius': transit_radii / petitRADTRANS.nat_cst.r_jup_mean
        },
        relative_tolerance=relative_tolerance
    )


def test_line_by_line_spectral_model_transmission_ccf():
    spectral_model = copy.deepcopy(spectral_model_lbl)

    spectral_model.model_parameters['temperature_profile_mode'] = 'isothermal'
    spectral_model.model_parameters['temperature'] = temperature_isothermal

    wavelengths, mock_transmission_data = spectral_model.get_spectrum_model(
        radtrans=radtrans_spectral_model_lbl,
        mode='transmission',
        parameters=None,
        update_parameters=True,
        telluric_transmittances=None,
        instrumental_deformations=None,
        noise_matrix=None,
        scale=True,
        shift=True,
        convolve=True,
        rebin=True,
        reduce=False
    )

    assert np.allclose(
        spectral_model.temperatures,
        temperature_isothermal,
        atol=0,
        rtol=relative_tolerance
    )

    for species in spectral_model.line_species:
        assert np.allclose(
            spectral_model.mass_mixing_ratios[species],
            radtrans_parameters['mass_fractions'][species],
            atol=0,
            rtol=relative_tolerance
        )

    assert np.allclose(
        spectral_model.mean_molar_masses,
        radtrans_parameters['mean_molar_mass'],
        atol=0,
        rtol=relative_tolerance
    )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['line_by_line_transmission_2d'],
        comparison_dict={
            'wavelengths': wavelengths,
            'relative_velocities': spectral_model.model_parameters['relative_velocities'],
            'transit_radii': mock_transmission_data
        },
        relative_tolerance=relative_tolerance
    )

    # Get models
    wavelengths_model, model = spectral_model.get_spectrum_model(
        radtrans=radtrans_spectral_model_lbl,
        mode='transmission',
        parameters=None,
        update_parameters=True,
        telluric_transmittances=None,
        instrumental_deformations=None,
        noise_matrix=None,
        scale=True,
        shift=True,
        convolve=True,
        rebin=False,
        reduce=False
    )

    # Cross-correlate spectrum with itself
    line_spread_function_fwhm = petitRADTRANS.nat_cst.c /\
        radtrans_parameters['mock_observation_parameters']['high_resolution_resolving_power']

    co_added_cross_correlations_snr, co_added_cross_correlations, \
        v_rest, kps, ccf_sum, ccfs, velocities_ccf, ccf_models, ccf_model_wavelengths \
        = petitRADTRANS.ccf.ccf.ccf_analysis(
            wavelengths_data=wavelengths,
            data=mock_transmission_data,
            wavelengths_model=wavelengths_model,
            model=model,
            velocities_ccf=None,
            model_velocities=spectral_model.model_parameters['relative_velocities'],
            normalize_ccf=radtrans_parameters['ccf_analysis_parameters']['normalize_ccf'],
            calculate_ccf_snr=radtrans_parameters['ccf_analysis_parameters']['calculate_ccf_snr'],
            ccf_sum_axes=radtrans_parameters['ccf_analysis_parameters']['ccf_sum_axes'],
            planet_radial_velocity_amplitude=spectral_model.model_parameters['planet_radial_velocity_amplitude'],
            system_observer_radial_velocities=spectral_model.model_parameters['system_observer_radial_velocities'],
            orbital_phases=spectral_model.model_parameters['orbital_phases'],
            planet_orbital_inclination=spectral_model.model_parameters['planet_orbital_inclination'],
            line_spread_function_fwhm=line_spread_function_fwhm,
            pixels_per_resolution_element=radtrans_parameters['ccf_analysis_parameters'][
                'pixels_per_resolution_element'],
            co_added_ccf_peak_width=line_spread_function_fwhm * radtrans_parameters[
                'ccf_analysis_parameters']['peak_lsf_factor'],
            velocity_interval_extension_factor=radtrans_parameters['ccf_analysis_parameters'][
                'velocity_interval_extension_factor'],
            kp_factor=radtrans_parameters['ccf_analysis_parameters']['kp_factor'],
            n_kp=None,
            n_vr=None,
            planet_radial_velocity_function=None
        )

    # Comparison
    compare_from_reference_file(
        reference_file=reference_filenames['co_added_cross_correlation'],
        comparison_dict={
            'rest_velocities': v_rest,
            'orbital_radial_velocity_amplitudes': kps,
            'co_added_cross_correlations': co_added_cross_correlations
        },
        relative_tolerance=relative_tolerance
    )



"""Wrapper of ESO's SKYCALC Command Line Interface.
"""

import copy
import json
import os
import tempfile

import numpy as np
from astropy.io import fits
from skycalc_cli.skycalc import AlmanacQuery, SkyModel

import petitRADTRANS.nat_cst as nc
from petitRADTRANS.utils import savez_compressed_record


def find_optimal_airmass_day(ra, dec, observatory='3060m',
                             time_step=nc.snc.day, time_range=nc.snc.Julian_year, start_time_mjd=6e4):
    modified_julian_dates = np.arange(0.0, time_range, time_step) / nc.snc.day + start_time_mjd

    best_airmass = np.inf
    best_time = -1
    airmass_pre = np.inf
    skip_backward = False

    # Forward search
    for i, mjd in enumerate(modified_julian_dates):
        print(
            f"Querying Almanac for (RA {ra} dec {dec}) "
            f"at {mjd} MJD ({i + 1}/{modified_julian_dates.size})...",
            end='\r'
        )

        try:
            airmass = get_airmass(
                ra=ra,
                dec=dec,
                observatory=observatory,
                mjd=mjd
            )
        except json.JSONDecodeError:
            print(f"Skipping invalid result")
            continue

        if 1.0 <= airmass < best_airmass:
            best_airmass = airmass
            best_time = mjd
        elif airmass > airmass_pre:
            if i != 0:
                skip_backward = True  # the peak airmass has been found, no need to run the backward search

            break
        elif airmass == -1 and np.inf > airmass_pre > -1:
            break

        airmass_pre = airmass

    airmass_pre = np.inf

    # Backward search
    if not skip_backward:
        for i, mjd in enumerate(modified_julian_dates[::-1]):
            print(
                f"Querying Almanac for (RA {ra} dec {dec}) "
                f"at {mjd} MJD ({modified_julian_dates.size - i}/{modified_julian_dates.size})...",
                end='\r'
            )

            try:
                airmass = get_airmass(
                    ra=ra,
                    dec=dec,
                    observatory=observatory,
                    mjd=mjd
                )
            except json.JSONDecodeError:
                print(f"Skipping invalid result")
                continue

            if 1.0 <= airmass < best_airmass:
                best_airmass = airmass
                best_time = mjd
            elif airmass > airmass_pre:
                break
            elif airmass == -1 and np.inf > airmass_pre > -1:
                break

            airmass_pre = airmass

    print('\nDone')

    return best_time


def get_airmass(ra, dec, mjd, observatory='3060m'):
    """Get the airmass of a sky point at a given date from a given observatory, using ESO's Almanac.

    Args:
        ra: (deg) right ascension
        dec: (deg) declination
        mjd: (day) modified Julian date
        observatory: point of observation ("paranal", "lasilla", "3060m")

    Returns:
        The airmass.
    """
    indic = {
        'ra': ra,
        'dec': dec,
        'observatory': observatory,
        'mjd': mjd
    }

    alm = AlmanacQuery(indic)
    dic = alm.query()

    return dic['airmass']


def get_optimal_airmass_curve(ra, dec, dit, observation_duration, query_dit=120.0, observatory='3060m'):
    """Get the most favourable airmass curve of a sky point from an observation point, using ESO's Almanac.
    First, the day of the year at which the target is best observed is searched for. This corresponds to the day at
    which the airmass of the target at UTC 0 is the lowest. Local time, day and night, the moon position etc. are not
    taken into account.
    Then, the lowest airmass of this day is searched for. The dates are adjusted so that the lowest airmass corresponds
    to the observation mid-time.
    Finally, the curve is queried from the Almanac at the requested dates around the observation mid-time. The requested
    dates use the interval query_dit. The queried curve is then interpolated on dates using the requested DIT.

    Small query DIT gives more accurate results, but a lot of queries (slow) are necessary.

    Args:
        ra: (deg) right ascension
        dec: (deg) declination
        dit: (s) interval of time between observations
        observation_duration: (s) total duration of observations
        query_dit: (s) interval of time between queries
        observatory: point of observation ("paranal", "lasilla", "3060m")

    Returns:
        The times (s) to lowest airmass and the corresponding airmasses.
    """
    best_day = find_optimal_airmass_day(
        ra=ra,
        dec=dec,
        observatory=observatory,
        time_step=nc.snc.day,
        time_range=nc.snc.Julian_year
    )

    if query_dit is None:
        query_dit = dit
    elif query_dit < dit:
        query_dit = dit

    n_dit = int(np.ceil(observation_duration / dit))

    modified_julian_dates = np.concatenate(
        (np.arange(0.0, n_dit * dit / 2, query_dit), [n_dit * dit / 2])
    ) / nc.snc.day
    modified_julian_dates = np.concatenate((-modified_julian_dates[:0:-1], modified_julian_dates)) + best_day
    central_index = int(modified_julian_dates.size / 2)
    shift = 0

    airmasses = -np.ones(modified_julian_dates.size)

    # Find lowest airmass of the day
    # Try after the best day date
    print("Searching for lowest airmass...")

    for i, mjd in enumerate(modified_julian_dates[central_index:]):
        airmasses[central_index + i] = get_airmass(
            ra=ra,
            dec=dec,
            observatory=observatory,
            mjd=mjd
        )

        if central_index + i - 1 > 0:
            if airmasses[central_index + i] > airmasses[central_index + i - 1] > -1:
                shift = i - 1
                break

    if shift == 0:  # lowest airmass is before requested best day
        for i, mjd in enumerate(modified_julian_dates[central_index - 1::-1]):
            airmasses[central_index - 1 - i] = get_airmass(
                ra=ra,
                dec=dec,
                observatory=observatory,
                mjd=mjd
            )

            if central_index - 1 - i > 0:
                if airmasses[central_index - 1 - i] > airmasses[central_index - i] > -1:
                    shift = -i
                    break

    # Shift the dates array center to the lowest airmass
    modified_julian_dates = modified_julian_dates[central_index + shift] - best_day + modified_julian_dates
    airmasses = np.roll(airmasses, -shift)

    # Ensure that rolled over airmasses are -1
    if shift < 0:
        airmasses[:-shift] = -1
    elif shift > 0:
        airmasses[-shift:] = -1

    # Get query airmass curve
    for i, mjd in enumerate(modified_julian_dates):
        print(
            f"Querying Almanac for (RA {ra} dec {dec}) "
            f"at {mjd} MJD ({i + 1}/{modified_julian_dates.size})...",
            end='\r'
        )

        if airmasses[i] != -1:
            continue

        airmasses[i] = get_airmass(
            ra=ra,
            dec=dec,
            observatory=observatory,
            mjd=mjd
        )

    print('\nDone')

    # Interpolate queried airmass curve to requested times
    if query_dit != dit:
        modified_julian_dates_dit = np.concatenate(
            (np.arange(0.0, n_dit * dit / 2, dit), [n_dit * dit / 2])
        ) / nc.snc.day
        modified_julian_dates_dit = np.concatenate(
            (-modified_julian_dates_dit[:0:-1], modified_julian_dates_dit)
        ) + modified_julian_dates[central_index]

        airmasses = np.interp(modified_julian_dates_dit, modified_julian_dates, airmasses)
    else:
        modified_julian_dates_dit = modified_julian_dates

    times = (modified_julian_dates_dit - modified_julian_dates[central_index]) * nc.snc.day

    return times, airmasses


def get_telluric_data(airmass=1.0, pwv_mode='pwv', season=0, time=0, pwv=3.5, msolflux=130.0, incl_moon='Y',
                      moon_sun_sep=90.0, moon_target_sep=45.0, moon_alt=45.0, moon_earth_dist=1.0,
                      incl_starlight='Y', incl_zodiacal='Y', ecl_lon=135.0, ecl_lat=90.0, incl_loweratm='Y',
                      incl_upperatm='Y', incl_airglow='Y', incl_therm='N', therm_t1=0.0, therm_e1=0.0,
                      therm_t2=0.0, therm_e2=0.0, therm_t3=0.0, therm_e3=0.0, vacair='vac', wmin=300.0,
                      wmax=2000.0, wgrid_mode='fixed_spectral_resolution', wdelta=0.1, wres=1000000,
                      wgrid_user=None, lsf_type='none', lsf_gauss_fwhm=5.0, lsf_boxcar_fwhm=5.0,
                      observatory='3060', temp_flag=0):
    if wgrid_user is None:
        wgrid_user = [500.0, 510.0, 520.0, 530.0, 540.0, 550.0]

    sky_model = SkyModel()

    params = {
        # Airmass. Alt and airmass are coupled through the plane parallel
        # approximation airmass=sec(z), z being the zenith distance
        # z=90°−Alt
        'airmass': airmass,  # float range [1.0,3.0]

        # Season and Period of Night
        'pwv_mode': pwv_mode,  # string  grid ['pwv','season']
        # integer grid [0,1,2,3,4,5,6] (0=all year, 1=dec/jan,2=feb/mar...)
        'season': season,
        # third of night integer grid [0,1,2,3] (0=all year, 1,2,3 = third
        # of night)
        'time': time,

        # Precipitable Water Vapor PWV
        # mm float grid [-1.0,0.5,1.0,1.5,2.5,3.5,5.0,7.5,10.0,20.0]
        'pwv': pwv,

        # Monthly Averaged Solar Flux
        'msolflux': msolflux,  # s.f.u float > 0

        # Scattered Moon Light
        # Moon coordinate constraints: |z – zmoon| ≤ ρ ≤ |z + zmoon| where
        # ρ=moon/target separation, z=90°−target altitude and
        # zmoon=90°−moon altitude.
        # string grid ['Y','N'] flag for inclusion of scattered moonlight.
        'incl_moon': incl_moon,
        # degrees float range [0.0,360.0] Separation of Sun and Moon as
        # seen from Earth ("moon phase")
        'moon_sun_sep': moon_sun_sep,
        # degrees float range [0.0,180.0] Moon-Target Separation ( ρ )
        'moon_target_sep': moon_target_sep,
        # degrees float range [-90.0,90.0] Moon Altitude over Horizon
        'moon_alt': moon_alt,
        # float range [0.91,1.08] Moon-Earth Distance (mean=1)
        'moon_earth_dist': moon_earth_dist,

        # Starlight
        # string  grid ['Y','N'] flag for inclusion of scattered starlight
        'incl_starlight': incl_starlight,

        # Zodiacal light
        # string grid ['Y','N'] flag for inclusion of zodiacal light
        'incl_zodiacal': incl_zodiacal,
        # degrees float range [-180.0,180.0] Heliocentric ecliptic
        # longitude
        'ecl_lon': ecl_lon,
        # degrees float range [-90.0,90.0] Ecliptic latitude
        'ecl_lat': ecl_lat,

        # Molecular Emission of Lower Atmosphere
        # string grid ['Y','N'] flag for inclusion of lower atmosphere
        'incl_loweratm': incl_loweratm,
        # Emission Lines of Upper Atmosphere
        # string grid ['Y','N'] flag for inclusion of upper stmosphere
        'incl_upperatm': incl_upperatm,
        # Airglow Continuum (Residual Continuum)
        # string grid ['Y','N'] flag for inclusion of airglow
        'incl_airglow': incl_airglow,

        # Instrumental Thermal Emission This radiance component represents
        # an instrumental effect. The emission is provided relative to the
        # other model components. To obtain the correct absolute flux, an
        # instrumental response curve must be applied to the resulting
        # model spectrum See section 6.2.4 in the documentation
        # http://localhost/observing/etc/doc/skycalc/
        # The_Cerro_Paranal_Advanced_Sky_Model.pdf
        # string grid ['Y','N'] flag for inclusion of instrumental thermal
        # radiation
        'incl_therm': incl_therm,
        'therm_t1': therm_t1,  # K float > 0
        'therm_e1': therm_e1,  # float range [0,1]
        'therm_t2': therm_t2,  # K float > 0
        'therm_e2': therm_e2,  # float range [0,1]
        'therm_t3': therm_t3,  # float > 0
        'therm_e3': therm_e3,  # K float range [0,1]

        # Wavelength Grid
        'vacair': vacair,  # vac or air
        'wmin': wmin,  # nm float range [300.0,30000.0] < wmax
        'wmax': wmax,  # nm float range [300.0,30000.0] > wmin
        # string grid ['fixed_spectral_resolution','fixed_wavelength_step',
        #              'user']
        'wgrid_mode': wgrid_mode,
        # nm/step float range [0,30000.0] wavelength sampling step dlam
        # (not the res.element)
        'wdelta': wdelta,
        # float range [0,1.0e6] RESOLUTION is misleading, it is rather
        # lam/dlam where dlam is wavelength step (not the res.element)
        'wres': wres,
        'wgrid_user': wgrid_user,
        # convolve by Line Spread Function
        'lsf_type': lsf_type,  # string grid ['none','Gaussian','Boxcar']
        'lsf_gauss_fwhm': lsf_gauss_fwhm,  # wavelength bins float > 0
        'lsf_boxcar_fwhm': lsf_boxcar_fwhm,  # wavelength bins float > 0
        'observatory': observatory,  # paranal
        # compute temperature and bolometric radiance
        'temp_flag': temp_flag
    }

    for key, value in params.items():
        sky_model.params[key] = value

    sky_model.call()

    # Create a temporary file to dump the data into and read it later
    # Data are fits-formatted and cannot be read as it, thus the data has to be saved into a .fits file first, then read
    # with io.fits
    # io.fits cannot work if the file is already opened for writing, so it needs to be:
    # created, written, closed, re-opened with io.fits and read, before it can be deleted
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(sky_model.data)

    with fits.open(f.name) as f_fits:
        data = np.array(copy.deepcopy(f_fits[1].data))

    # Delete the temporary file
    os.unlink(f.name)

    return data


def get_tellurics_npz(file, wavelength_range=None, rewrite=False, **kwargs):
    """Get telluric transmittance and wavelength from a .npz file.
    If the file doesn't exist, a request to the skycalc server is sent to retrieve the data.
    While all the skycalc information is saved into the file, this function extracts only the wavelengths and
    transmittances.
    To prevent issues with large wavelength ranges, the request is split into 2.
    # TODO splitting could be much smarter (no split for small ranges, multiple splits for very large ones)

    Args:
        file: file from which to load the telluric transmittance and corresponding wavelengths
        wavelength_range: (um) list containing the min and max wavelengths
        rewrite: if True, the file is rewritten even if it already exists
        **kwargs: get_telluric_data arguments

    Returns:
        The wavelengths (um) and corresponding telluric transmittances
    """
    if not os.path.isfile(file) or rewrite:
        if wavelength_range is None:
            raise ValueError(f"argument 'wavelength_range' cannot be None: file '{file}' do not exists")

        # Prevent the CLI to complain about requesting too much data by splitting the request into 2
        wavelength_range_ = copy.deepcopy(wavelength_range)
        wavelength_range_[-1] = np.diff(wavelength_range)[0] / 2 + wavelength_range[0]

        telluric_data0 = get_telluric_data(
            wmin=wavelength_range_[0] * 1e3,  # um to nm
            wmax=wavelength_range_[1] * 1e3,  # um to nm
            **kwargs
        )

        wavelength_range_[0] = np.max(telluric_data0['lam'][-2]) * 1e-3
        wavelength_range_[-1] = wavelength_range[-1]

        telluric_data1 = get_telluric_data(
            wmin=wavelength_range_[0] * 1e3,  # um to nm
            wmax=wavelength_range_[1] * 1e3,  # um to nm
            **kwargs
        )

        wh = np.greater(telluric_data1['lam'], np.max(telluric_data0['lam']))
        wavelength_range_ = np.concatenate((
                telluric_data0['lam'],
                telluric_data1['lam'][wh]
            ))
        telluric_data = np.recarray(wavelength_range_.shape, dtype=telluric_data0.dtype)

        for key in telluric_data.dtype.names:
            telluric_data[key] = np.concatenate((
                telluric_data0[key],
                telluric_data1[key][wh]
            ))

        savez_compressed_record(file, telluric_data)
    else:
        telluric_data = np.load(file)

    wavelengths_telluric = telluric_data['lam'] * 1e-3  # nm to um
    telluric_transmittance = telluric_data['trans']

    return wavelengths_telluric, telluric_transmittance
